import torch
class CNNscheduler:
        def __init__(self, optimizer) -> None:
            self.scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, 0.5)
        def update(self, epoch):
            if (epoch > 100) & (epoch % 10 == 9):
                self.scheduler.step()
        def load_state_dict(self, state_dict):
            self.scheduler.load_state_dict(state_dict)
        def state_dict(self):
            return self.scheduler.state_dict()
        def get_last_lr(self):
            return self.scheduler.get_last_lr()
class SNNscheduler(CNNscheduler):
    def __init__(self, optimizer) -> None:
        super(SNNscheduler, self).__init__(optimizer)
        self.scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, [30, 60, 80, 100], gamma=0.2)
    def update(self, epoch):
        self.scheduler.step()

optimizer = torch.optim.Adam(torch.nn.Conv2d(2, 3,1,1).parameters(), lr=5e-4, weight_decay=5e-4)
lr_scheduler = SNNscheduler(optimizer)
for epoch in range(150):
    print(epoch, lr_scheduler.get_last_lr())
    lr_scheduler.update(epoch)


import numpy as np
import os
import struct
import random

__all__ = ["event", "loadaerdat", "fromArray", "load"]



class event:
    def __init__(self, p, x, y, t) -> None:
        assert len(x) == len(y) and len(x) == len(p) and len(x) == len(t), "pxyt must be the same length."
        self.length = len(x)
        self.p, self.x, self.y, self.t = np.array(p), np.array(x), np.array(y), np.array(t)
    
    def keys(self):
        return ['p', 'x', 'y', 't']

    def __len__(self):
        '''
        return the len of the event
        '''
        return self.length
    
    def __getitem__(self, key):
        if isinstance(key, str):
            if key is 'p':
                return self.p
            elif key is 'x':
                return self.x
            elif key is 'y':
                return self.y
            elif key is 't':
                return self.t
            else:
                print(f"you can just extract p, x, y, t. but get {key}")
        elif isinstance(key, int):
            if key>=0 and key < self.len():
                return (self.p[key], self.x[key], self.y[key], self.t[key])
            else:
                raise IndexError("index out of range.")
        raise ValueError("unsupport extract.")
    
    def __iter__(self):
        '''
        iter the events and return a generator, event will be (p, x, y, t)
        '''
        for i in range(self.len()):
            yield (self.p[i], self.x[i], self.y[i], self.t[i])

    def __eq__(self, __o: object) -> bool:
        '''
        compare if two event are the same. 
        return True if they are same, otherwise False.
        '''
        assert isinstance(__o, event), 'event only can be compared with event.'
        return (self.x == __o.x).all() and (self.y == __o.y).all() and \
                (self.p == __o.p).all() and (self.t == __o.p).all()
    
    def len(self):
        return self.__len__()
    
    def size(self):
        return self.len()*4

    def shape(self):
        return (self.len(), 4)

    def checkValid(self, dim_p:int=-1, dim_x:int=-1, dim_y:int=-1, dim_t:int=-1):
        return len(self.getValidIndex(dim_p, dim_x, dim_y, dim_t)) == self.len()
        
    def getValidIndex(self, dim_p:int=-1, dim_x:int=-1, dim_y:int=-1, dim_t:int=-1):
        '''
        get the index  where pxyt >=0 and < dim
        '''
        index_p = (self.p >= 0)
        if dim_p > 0:
            index_p &= (self.p < dim_p)

        index_x = (self.x >= 0)
        if dim_x > 0:
            index_x &= (self.x < dim_x)

        index_y = (self.y >= 0)
        if dim_y > 0:
            index_y &= (self.y < dim_y)

        index_t = (self.t >= 0)
        if dim_t > 0:
            index_t &= (self.t < dim_t)
        valid = np.argwhere( index_p & index_x & index_y & index_t)
        return valid.reshape(-1)
    
    def _removeInvalid(self, dim_p:int=-1, dim_x:int=-1, dim_y:int=-1, dim_t:int=-1)->None:
        '''
        in-place method, keep the valid index. (0<=pxyt<dim)
        '''
        valid = self.getValidIndex(dim_p, dim_x, dim_y, dim_t)
        self.p = self.p[valid]
        self.x = self.x[valid]
        self.y = self.y[valid]
        self.t = self.t[valid]

    def removeInvalid(self, dim_p:int=-1, dim_x:int=-1, dim_y:int=-1, dim_t:int=-1):
        '''
        remove the invalid index. (0<=pxyt<dim)
        it will not change the event, but return a required event.
        '''
        valid = self.getValidIndex(dim_p, dim_x, dim_y, dim_t)
        return event(self.p[valid], self.x[valid], self.y[valid], self.t[valid])

    def _toMillisecond(self)->None:
        '''
        it will turn event's timestamp from us to ms. 
        in-place modify and will not return anything.
        '''
        self.t /= 1000
    
    def toMillisecond(self):
        '''
        turn event's timestamp from us to ms. 
        it will not change the event, but return a required event.
        '''
        return event(self.p, self.x, self.y, self.t/1000)
    
    def toArray(self, order:str="pxyt"):
        assert len(order) == 4 and order[0] in self.keys() and order[1] in self.keys() \
            and order[2] in self.keys() and order[3] in self.keys(), "order is a string consist of p, x, y, t with length 4."
        return np.column_stack((self[order[0]], self[order[1]], self[order[2]], self[order[3]]))

    def toTimeStep(self, dim:tuple, way:str="count", reshape=False)->np.ndarray:
        """
        convert events to C H W T. which mean handle events with oclock driven.
        And aggregates events at the same time interval into a step frame.
        @param:dim. the output shape of ndarray, which is a tuple with length 4.
        @param:way. how the timestamps aggregates to step.
        @param:reshape. if the Width and Height in dim are smaller than events' resolution.
        """
        if reshape:
            self._removeInvalid(dim_p=dim[0], dim_x=dim[2], dim_y=dim[1])
        t = (dim[3]*(self.t / (max(self.t)+1))).astype(int)
        if way == "count":
            steps = np.zeros(dim).reshape(-1)
            index = np.ravel_multi_index((self.p, self.y, self.x, t), dim)
            weight_index = np.bincount(index)
            steps[:weight_index.size] = weight_index
            steps = steps.reshape(dim)
            return steps
    
    def domainReverse(self, domain, HW, filter_events):
        '''
        reverse specific events(filter_events) in the whole domain.
        of course you can reverse specific events in specific domain, but I don't find if it is meaningful now. so i havn't implement it now.
        '''
        # assert domain in self.keys(), f"you can just extract p, x, y, t. but get {domain}."
        
        if domain is 'p':
            self.p[filter_events] = 1-self.p[filter_events]
        elif domain is 'x':
            self.x[filter_events] = HW[1]-1-self.x[filter_events]
        elif domain is 'y':
            self.y[filter_events] = HW[0]-1-self.y[filter_events]
        elif domain is 't': 
            self.t[filter_events] = max(self.t) -self.t[filter_events]  
        
    def domainDrift(self, domain, HW, filter_events, distance, recurrent=False):
        '''
        drift specific events range[begin, end) in the whole domain.
        of course you can drift specific events in specific domain, but I don't find if it is meaningful now. so i havn't implement it now.
        '''
        # assert domain in self.keys(), f"you can just extract p, x, y, t. but get {domain}."
        
        if domain is 'p':
            distance = np.random.randint(-int(distance*2), int(distance*2)+1) 
            self.p[filter_events] = self.p[filter_events] + distance
            if recurrent:
                self.p[filter_events] %= 2
            else:
                self._removeInvalid(dim_p=2)
        elif domain is 'x':
            distance = np.random.randint(-int(distance*HW[1]), int(distance*HW[1])+1) 
            self.x[filter_events] = self.x[filter_events] + distance
            if recurrent:
                self.x[filter_events] %= HW[1]
            else:
                self._removeInvalid(dim_x=HW[1])
        elif domain is 'y':
            distance = np.random.randint(-int(distance*HW[0]), int(distance*HW[0])+1) 
            self.y[filter_events] = self.y[filter_events] + distance
            if recurrent:
                self.y[filter_events] %= HW[0]
            else:
                self._removeInvalid(dim_y=HW[0])
        elif domain is 't':
            resolution = max(self.t)+1
            distance = np.random.randint(-int(distance*resolution), int(distance*resolution)+1) 
            self.t[filter_events] = self.t[filter_events] + distance
            if recurrent:
                self.t[filter_events] %= resolution
            else:
                self._removeInvalid(dim_t=resolution)  

    def event_drop(self, HW):
        x = self.x
        y = self.y
        t = self.t
        p = self.p
        option = np.random.randint(0, 4) # 0: identity, 1: drop_by_time, 2: drop_by_area, 3: random_drop
        if option == 1: # drop_by_time
            T = np.random.randint(1, 10)/10.0 #np.random.uniform(0.1, 0.9)
            self.drop_by_time(T=T)
        elif option == 2: # drop by area
            area_ratio = np.random.randint(1, 6) / 20.0 #np.random.uniform(0.05, 0.1, 0.15, 0.2, 0.25)
            self.drop_by_area(resolution=HW, area_ratio=area_ratio)
        elif option == 3: # random drop
            ratio = np.random.randint(1, 10)/10.0 #np.random.uniform(0.1, 0.9)
            self.random_drop(ratio=ratio)

        if len(self.x) == 0:  # avoid dropping all the events
            self.p = p
            self.x = x
            self.y = y
            self.t = t

    def drop_by_time(self,T= 0):
        # assert 0.1 <= T <= 0.5

        # time interval
        t_start = np.random.uniform(0, 1)
        if T == 0: # randomly choose a value between [0.1, 0.9]
            T = np.random.randint(1, 10)/10.0
        t_end = t_start + T

        timestamps = self.t
        max_t = max(timestamps) 
        idx = (timestamps < (max_t*t_start)) | (timestamps > (max_t*t_end)) # remaining events that are not within the given time interval

        self.p = self.p[idx]
        self.x = self.x[idx]
        self.y = self.y[idx]
        self.t = self.t[idx]

    def drop_by_area(self,resolution, area_ratio=0):
        # assert 0.1 <= area_ratio <= 0.3 

        # get the area whose events are to be dropped
        x0 = np.random.uniform(resolution[0])
        y0 = np.random.uniform(resolution[1])

        if area_ratio == 0:
            area_ratio = np.random.randint(1, 6)/20.0

        x_out = resolution[0] * area_ratio
        y_out = resolution[1] * area_ratio

        x0 = int(max(0, x0 - x_out/2.0))
        y0 = int(max(0, y0 - y_out/2.0))

        x1 = min(resolution[0], x0 + x_out)
        y1 = min(resolution[1], y0 + y_out)

        xy = (x0, x1, y0, y1) # rectangele to be dropped

        idx1 = (self.x < xy[0]) | (self.x > xy[1])
        idx2 = (self.y < xy[2]) | (self.y > xy[3])
        idx = idx1 & idx2

        self.p = self.p[idx]
        self.x = self.x[idx]
        self.y = self.y[idx]
        self.t = self.t[idx]

    def random_drop(self,ratio = 0):
        # assert 0.1 <= ratio <= 0.5

        if ratio == 0:
            ratio = np.random.randint(1, 10)/10.0

        N = len(self.x) # number of total events
        num_drop = int(N * ratio) # number of events to be dropped
        idx = random.sample(list(np.arange(0, N)), N-num_drop)

        self.p = self.p[idx]
        self.x = self.x[idx]
        self.y = self.y[idx]
        self.t = self.t[idx]


    def save(self, path, extention):
        """
        save event to Files with the suffix "extention".
        """
        assert extention in [".npy", ".mat", "npy", "mat"], f"event only support to save to numpy or matlab files, get extention:{extention}"
        if extention in ['.npy', 'npy']:
            np.save(path, self.toArray())
        else:
            from scipy.io import savemat
            savemat(path, {'item':self.toArray()}) # if ndarray cannot be save to mat, use tolist()

    def saveaerdat(self, path):
        """
        save to event_format. not implement till now.
        """
        raise NotImplementedError

def fromArray(data:np.ndarray)->event:
    assert data.shape[1] == 4, "data must be a ndarray with shape (..., 4)."
    return event(data[:,0], data[:,1], data[:,2], data[:,3])  

def save(file:str, obj:event, extention:str):
    obj.save(file, extention)

def load(filename, extention)->event:
    """
    load event from Files with the suffix "extention".
    """
    assert extention in [".npy", ".mat", "npy", "mat"], f"event only support to load from numpy or matlab files, get extention:{extention}"
    if extention in ['.npy', 'npy']:
        data = np.load(filename, allow_pickle=True)
    else:
        import scipy.io as scio
        data = scio.loadmat(filename, verify_compressed_data_integrity=False)['item'].astype(np.int64)
    return fromArray(data)

def loadaerdat(filename='/tmp/aerout.dat', camera='DVS128')->event:
    """    
    This function read events .
    load AER data file and parse these properties of AE events:
    - timestamps (in us), 
    - x,y-position [0..127]
    - polarity (0/1)
    @param filename - path to the file to read
    @param length - how many bytes(B) should be read; default 0=whole file
    @param version - which file format version is used: "aedat" = v2, "dat" = v1 (old)
    @param debug - 0 = silent, 1 (default) = print summary, >=2 = print all debug
    @param camera='DVS128' or 'DAVIS240'
    @return (ts, xpos, ypos, pol) 4-tuple of lists containing data of all events;
    """


    # camera
    if(camera == 'DVS128'):
        xmask = 0x00fe
        xshift = 1
        ymask = 0x7f00
        yshift = 8
        pmask = 0x1
        pshift = 0
        aeLen = 8
        readMode = '>II'

    elif(camera == 'DAVIS240'):
        xmask = 0x003ff000
        xshift = 12
        ymask = 0x7fc00000
        yshift = 22
        pmask = 0x800
        pshift = 11
        eventtypeshift = 31

    elif(camera == 'N'):
        xmask = 0x3fff
        xshift = 0
        ymask = 0xfffc000
        yshift = 14
        pmask = 0x10000000
        pshift = 28
        aeLen = 8 #useless
        readMode = [('t', 'u4'), ('_', 'i4')]
    elif(camera == 'N101'):
        f = open(filename, 'rb')
        raw_data = np.fromfile(f, dtype=np.uint8)
        f.close()
        raw_data = np.uint32(raw_data)
        xaddr = raw_data[0::5]
        yaddr = raw_data[1::5]
        pol = (raw_data[2::5] & 128) >> 7 #bit 7
        timestamps = ((raw_data[2::5] & 127) << 16) | (raw_data[3::5] << 8) | (raw_data[4::5])
        return event(pol, xaddr, yaddr, timestamps)
    else:
        raise ValueError("Unsupported camera: %s" % (camera))

    # find header
    aerdatafh = open(filename, 'rb')
    k = 0  # line number
    p = 0  # pointer, position on bytes
    statinfo = os.stat(filename)
    if camera == 'N':
        # This code comes from NCARS Datasets.
        lt = aerdatafh.readline()
        while lt and lt.decode("latin-1")[:2] == '% ':
            p += len(lt)
            k += 1
            lt = aerdatafh.readline()
        ev_type = np.frombuffer(aerdatafh.read(1), dtype=np.uint8)[0]
        p += 1
        ev_size = np.frombuffer(aerdatafh.read(1), dtype=np.uint8)[0]
        p += 1

    elif camera == 'DVS128':
        lt = aerdatafh.readline()
        while lt and chr(lt[0]) == "#":
            p += len(lt)
            k += 1
            lt = aerdatafh.readline()


    # conversion
    aerdatafh.seek(p)
    if camera == 'N':
        dat = np.fromfile(aerdatafh, dtype=readMode, count=-1)
        if ('_', 'i4') in readMode:
            timestamps = dat["t"].tolist()
            xaddr = np.bitwise_and(dat["_"], xmask).tolist()
            yaddr = np.right_shift(np.bitwise_and(dat["_"], ymask), yshift).tolist()
            pol = np.right_shift(np.bitwise_and(dat["_"], pmask), pshift).tolist()
            
    elif camera == 'DVS128':
        length = statinfo.st_size
        timestamps = []
        xaddr = []
        yaddr = []
        pol = []
        while p < length:
            # read data-part of file
            aerdatafh.seek(p)
            s = aerdatafh.read(aeLen)
            p += aeLen
            addr, ts = struct.unpack(readMode, s)
            x_addr = (addr & xmask) >> xshift
            y_addr = (addr & ymask) >> yshift
            a_pol = (addr & pmask) >> pshift

            timestamps.append(ts)
            xaddr.append(x_addr)
            yaddr.append(y_addr)
            pol.append(a_pol)

    return event(pol, xaddr, yaddr, timestamps)



import torch
import torch.nn as nn

class Neuron(nn.Module):
    def __init__(self, response_func, threshold, leak, _scale=1, pool=False, out=False, name="neuron") -> None:
        super(Neuron, self).__init__()
        self.response_func = response_func
        self.threshold = threshold
        self.leak = leak
        self.name = name
        self._scale = torch.nn.parameter.Parameter(torch.tensor(_scale), requires_grad=False)
        self.pool = pool
        self.out = out
    
    def reset(self, soma_in, mem, soma_out):
        # torch.autograd.Variable(torch.zeros((batch_size, *self.in_shape), device=device), requires_grad=False)  # will not add in optimizer
        self.soma_in = soma_in
        self.mem = mem
        self.soma_out = soma_out

    def snn_forward(self, response_func_input):
        soma_in = self.response_func(response_func_input)
        self.soma_in += soma_in
        self.mem += soma_in
        # 后面再修改threshold， gt成一个函数
        ex_membrane = nn.functional.threshold(self.mem, self.threshold, 0)
        self.mem = self.mem - ex_membrane # hard reset
        # generate spike
        soma_out = ex_membrane.gt(0).type(torch.cuda.FloatTensor)
        if not self.pool:
            self.mem = self.leak * self.mem
        self.soma_out += soma_out
        return soma_out

    def forward(self, sum_x):
        sum_y = self.response_func(sum_x)
        out = torch.mul(sum_y, self.scale)
        return out - out.detach() + self.soma_out.detach()

    def updateScale(self, sign):
        if sign:
            index = self.soma_in >= self.soma_out
            new_scale = self.soma_out[index].sum()/self.soma_in[index].sum() if self.soma_in[index].sum() > 1e-3 else 1
            self._scale.data = 0.6 * new_scale + 0.4 * self._scale.data
        self.scale = self._scale.data * self.soma_in.gt(0).type(torch.cuda.FloatTensor)

        


import sys
sys.path.append("/home/shb/SpikingNeuralNetwork/") 

from SNN import event as ev
from SNN.utils import transformer
import numpy as np
print(np.random.randint(1))
a = [1,2,3,4,5]
a[2:] = [5,6,7]
t = ev.loadaerdat("/home/shb/datasets/CIFAR10DVS/extract/airplane/cifar10_airplane_0.aedat")
print(t['x'])
print(t['y'])
t1 = transformer.EventReverse(transformer.EventReverse.x)(t)
print(t1['x'])
print(t1['y'])


from .layer import QuantizationLayerEST, QuantizationLayerEventCount, QuantizationLayerEventFrame, QuantizationLayerVoxGrid
from .layer import ALTP_block, ALTP_D
from .node import Neuron
from .vgg import vgg19, vgg7, snnvgg7, snnvgg19
from .resnet import resnet9, resnet18, resnet34, snnresnet9, snnresnet18
from torchvision.models.mobilenet import mobilenet_v2

import math
import torch
import torch.nn as nn

def Classifier(
    representation:str, 
    model_name:str, 
    img_dim:tuple, 
    in_channel:int,
    num_class:int, 
    repeat:int=1, 
    avgtimes:int=1, 
    vth:int=1, 
    leak:int=1, 
    pretrained:bool=True) -> nn.Module:
    '''
    build a model with arch 'model_name', the data will be handled by representation method.
    img_dim is not include channel(default 2 for event data).
    Specifically, for eventcount and eventframe, the img_dim is (H, W)
    for est and voxgrid, the img_dim is (voxel, H, W)
    for timeSteps, the img_dim is (H, W, T)
    The repeat and avgtimes is used for reproduce the AAAI2021 paper, 
    which can be found at https://www.aaai.org/AAAI21Papers/AAAI-4138.WuHao.pdf
    we adapt this paper as SNN's baseline.
    '''
    representation = representation.lower()
    model_name = model_name.lower()
    if representation == 'timesteps':
        return SNNClassifier(img_dim, num_class, model_name, avgtimes, repeat, vth, leak)
    elif representation in ['est', 'voxgrid', 'eventcount', 'eventframe']:
        return CNNClassifier(img_dim, num_class, in_channel, representation, model_name, pretrained)


        

class CNNClassifier(nn.Module):
    def __init__(self,
                 voxel_dimension=(9,128,128),  # dimension of voxel will be C x 2 x H x W
                 num_classes=10,
                 in_channel=2,
                 representation='est', #'EventCount', 'VoxGrid', 'EventFrame'
                 model_name = 'vgg19',
                 pretrained=True):

        nn.Module.__init__(self)
        
        if representation == 'eventcount':
            self.quantization_layer = QuantizationLayerEventCount(voxel_dimension)
        elif representation == 'eventframe':
            self.quantization_layer = QuantizationLayerEventFrame(voxel_dimension)
        elif representation == 'voxgrid':
            self.quantization_layer = QuantizationLayerVoxGrid(voxel_dimension)
        elif representation == 'est':
            self.quantization_layer = QuantizationLayerEST(voxel_dimension)

        # replace fc layer and first convolutional layer
        input_channels = in_channel # for event count and event frame
        if len(voxel_dimension) == 3: # for vox grid and EST
            input_channels *= voxel_dimension[0]

        if model_name == 'vgg7':
            self.classifier = vgg7(pretrained=pretrained)
            self.classifier.features[0] = nn.Conv2d(input_channels, 64, kernel_size=3, padding=1)
            self.classifier.classifier[-1] = nn.Linear(self.classifier.classifier[-1].in_features, num_classes)
            # self.classifier.classifier[0] = nn.Linear(256*int(voxel_dimension[-1]/4)*int(voxel_dimension[-2]/4), 1024, bias=False)
        elif model_name == 'resnet9':
            self.classifier = resnet9(pretrained=pretrained)
            self.classifier.features[0] = nn.Conv2d(input_channels, 64, kernel_size=3, stride=1, padding=1, bias=False)
            self.classifier.classifier.fc2 = nn.Linear(self.classifier.classifier.fc2.in_features, num_classes)
            # self.classifier.classifier.fc2 = nn.Linear(512*int((voxel_dimension[-1]-2)/8)*int((voxel_dimension[-2]-2)/8), num_classes, bias=False)
        elif model_name == 'vgg19':
            self.classifier = vgg19(pretrained=pretrained)   # nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
            self.classifier.features[0] = nn.Conv2d(input_channels, 64, kernel_size=3, padding=1)
            self.classifier.classifier[-1] = nn.Linear(4096, num_classes, bias=True)
        elif model_name == 'resnet18':
            self.classifier = resnet18(pretrained=pretrained)
            self.classifier.conv1 = nn.Conv2d(input_channels, 64, kernel_size=7, stride=2, padding=3, bias=False)
            self.classifier.fc = nn.Linear(self.classifier.fc.in_features, num_classes)
        elif model_name == 'resnet34':
            self.classifier = resnet34(pretrained=pretrained)
            self.classifier.conv1 = nn.Conv2d(input_channels, 64, kernel_size=7, stride=2, padding=3, bias=False)
            self.classifier.fc = nn.Linear(self.classifier.fc.in_features, num_classes)
        elif model_name == 'mobilenet_v2':
            self.classifier = mobilenet_v2(pretrained=pretrained)
            self.classifier.features[0][0] = nn.Conv2d(input_channels, 32, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1), bias=False)
            self.classifier.classifier[-1] = nn.Linear(in_features=1280, out_features=num_classes, bias=True)

    def forward(self, x):
        vox = self.quantization_layer.forward(x).float()
        pred = self.classifier.forward(vox)
        return pred

class SNNClassifier(nn.Module):
    def __init__(self,
        img_dim,
        num_classes,
        model_name:str,
        avgtimes:int=1,
        repeat:int=1,
        vth:int=1,
        leak:int=1
        ) -> None:

        super(SNNClassifier, self).__init__()
        
        self.steps = img_dim[-1]
        self.repeat = repeat
        self.avgtimes = avgtimes
        self.downsample = nn.AvgPool2d(kernel_size=3, stride=3)
        if model_name.lower() == 'vgg7':
            self.arch = snnvgg7(vth, leak, num_classes, (2, int(img_dim[0]/3), int(img_dim[1]/3)))           
        elif model_name.lower() == 'resnet9':
            self.arch = snnresnet9(vth, leak, num_classes, (2, int(img_dim[0]/3), int(img_dim[1]/3)))    
        elif model_name.lower() == 'vgg19':
            self.arch = snnvgg19(vth, leak, num_classes, (2, int(img_dim[0]/3), int(img_dim[1]/3)))           
        elif model_name.lower() == 'resnet18':
            self.arch = snnresnet18(vth, leak, num_classes, (2, int(img_dim[0]/3), int(img_dim[1]/3)))        
        
        self._weight_init()
    
    def forward(self, x, updateSign=None):
        self.arch.reset_mem()
        total_input = 0
        with torch.no_grad():
            for i in range(self.steps * self.repeat):
                # Poisson input spike generation
                eventframe_input = x[:, :, :, :, i % self.steps].float()
                eventframe_input = self.avgtimes * self.downsample(eventframe_input)
                total_input += eventframe_input
                self.arch.snn_forward(eventframe_input)

            self.arch.updateScale(updateSign)
        
        out = self.arch(total_input)
        return out / self.steps

    def _weight_init(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.in_channels
                variance1 = math.sqrt(2.0 / n)
                m.weight.data.normal_(0, variance1)

            elif isinstance(m, nn.Linear):
                size = m.weight.size()
                fan_in = size[1]  # number of columns
                variance2 = math.sqrt(2.0 / fan_in)
                m.weight.data.normal_(0.0, variance2)

class SNN(nn.Module):
    def __init__(self, size, repeat, avgtimes, sign, device, vth, leak) -> None:
        super(SNN, self).__init__()
        self.repeat = repeat
        self.avgtimes = avgtimes
        self.sign = sign
        self.vth = vth
        self.leak = leak
        H,W = size # list:[H,W]
        para_size_H = H
        para_size_W = W
        device = torch.device(f'cuda:{device}')
        self.maxpool4 = nn.AvgPool2d(kernel_size=3)
        para_size_H = int(para_size_H/3)
        para_size_W = int(para_size_W/3)
        conv11 = nn.Conv2d(in_channels=2, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron11 = Neuron(conv11, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv12 = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron12 = Neuron(conv12, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        avgpool1 = nn.AvgPool2d(kernel_size=2)
        self.neuronp1 = Neuron(avgpool1, 0.75, self.leak, pool=True)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        conv21 = nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron21 = Neuron(conv21, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv22 = nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron22 = Neuron(conv22, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv23 = nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron23 = Neuron(conv23, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.avgpool2 = nn.MaxPool2d(kernel_size=2)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)

        fc0 = nn.Linear(256 * para_size_H * para_size_W, 1024, bias=False)
        self.neuronf0 = Neuron(fc0, self.vth, self.leak)
        self.classifier = nn.Linear(1024, 10, bias=False)
        self.classifier.threshold = self.vth
        
        self._weight_init()

    def forward(self, input_x):
        self.reset_mem()
        total_input = 0
        with torch.no_grad():
            for i in range(self.steps * self.repeat):
                # Poisson input spike generation
                eventframe_input = input_x[:, :, :, :, i % self.steps].float()
                eventframe_input = self.avgtimes * self.maxpool4(eventframe_input)
                total_input += eventframe_input
                spike = self.neuron11.snn_forward(eventframe_input)
                spike = self.neuron12.snn_forward(spike)
                spike = self.neuronp1.snn_forward(spike)
                spike = self.neuron21.snn_forward(spike)
                spike = self.neuron22.snn_forward(spike)
                spike = self.neuron23.snn_forward(spike)

                fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

                spike = self.neuronf0.snn_forward(fc_in)
            self.updateScale(self.sign)
        with torch.enable_grad():
            spike = self.neuron11(total_input)
            spike = self.neuron12(spike)
            spike = self.neuronp1(spike)
            spike = self.neuron21(spike)
            spike = self.neuron22(spike)
            spike = self.neuron23(spike)

            fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

            character = self.neuronf0(fc_in)
            out = self.classifier(character)
        return out / self.classifier.threshold / self.steps
    

        

    def updateScale(self, sign):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.updateScale(sign)

    def reset_mem(self):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.reset(0, 0, 0)
        
    def _weight_init(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.in_channels
                variance1 = math.sqrt(2.0 / n)
                m.weight.data.normal_(0, variance1)

            elif isinstance(m, nn.Linear):
                size = m.weight.size()
                fan_in = size[1]  # number of columns
                variance2 = math.sqrt(2.0 / fan_in)
                m.weight.data.normal_(0.0, variance2)


class SNN1(nn.Module):
    def __init__(self, args) -> None:
        self.args = args
        super(SNN1, self).__init__()
        device = torch.device(f'cuda:{self.args.device}')
        # maxpool
        self.maxpool1 = nn.AvgPool2d(kernel_size=3)
        # conv1
        self.conv1 = ALTP_block(block_type='conv',conv_or_fc_in_channels=2,conv_or_fc_out_channels=64,conv_or_pool_kernel_size=3,
        conv_or_pool_stride=1,conv_or_pool_padding=1,conv_or_fc_in_bias=False,Neuron_threshold=self.self.vth,Neuron_leak=self.self.leak)
        # conv2
        self.conv2 = ALTP_block(block_type='conv',conv_or_fc_in_channels=64,conv_or_fc_out_channels=64,conv_or_pool_kernel_size=3,
        conv_or_pool_stride=1,conv_or_pool_padding=1,conv_or_fc_in_bias=False,Neuron_threshold=self.self.vth,Neuron_leak=self.self.leak)
        # avgpool
        self.avgpool1 = ALTP_block(block_type='avgpool',conv_or_pool_kernel_size=2,Neuron_threshold=0.75,Neuron_leak=self.self.leak,Neuron_pool=True)
        # conv3
        self.conv3 = ALTP_block(block_type='conv',conv_or_fc_in_channels=64,conv_or_fc_out_channels=128,conv_or_pool_kernel_size=3,
        conv_or_pool_stride=1,conv_or_pool_padding=1,conv_or_fc_in_bias=False,Neuron_threshold=self.self.vth,Neuron_leak=self.self.leak)
        # conv4
        self.conv4 = ALTP_block(block_type='conv',conv_or_fc_in_channels=128,conv_or_fc_out_channels=128,conv_or_pool_kernel_size=3,
        conv_or_pool_stride=1,conv_or_pool_padding=1,conv_or_fc_in_bias=False,Neuron_threshold=self.self.vth,Neuron_leak=self.self.leak)
        # conv5
        self.conv5 = ALTP_block(block_type='conv',conv_or_fc_in_channels=128,conv_or_fc_out_channels=256,conv_or_pool_kernel_size=3,
        conv_or_pool_stride=1,conv_or_pool_padding=1,conv_or_fc_in_bias=False,Neuron_threshold=self.self.vth,Neuron_leak=self.self.leak)
        # avgpool
        self.maxpool2 = nn.MaxPool2d(kernel_size=2)
        # fc1000
        self.fc1000 = ALTP_block(block_type='fc',conv_or_fc_in_channels=256*10*10,conv_or_fc_out_channels=1024,add_ALTP=True,ALTP_shape=(1024),
        ALTP_device=device,Neuron_threshold=self.self.vth,Neuron_leak=self.self.leak)
        # fc10
        self.classifier = nn.Linear(1024, 10, bias=False)
        self.classifier.threshold = self.self.vth
        
        self._weight_init()

    def forward(self, input_x, steps, l=1):
        self.reset_mem()
        total_input = 0
        with torch.no_grad():
            for i in range(steps * self.self.repeat):
                # Poisson input spike generation
                eventframe_input = input_x[:, :, :, :, i % steps].float()
                eventframe_input = self.self.avgtimes * self.maxpool1(eventframe_input)
                total_input += eventframe_input
                spike = self.conv1.snn_forward(eventframe_input)
                spike = self.conv2.snn_forward(spike)
                spike = self.avgpool1.snn_forward(spike)
                spike = self.conv3.snn_forward(spike)
                spike = self.conv4.snn_forward(spike)
                spike = self.conv5.snn_forward(spike)
                fc_in = self.maxpool2(spike).view(input_x.size(0), -1)
                spike = self.fc1000.snn_forward(fc_in)
            self.updateScale(self.self.sign)
        with torch.enable_grad():
            spike = self.conv1(total_input)
            spike = self.conv2(spike)
            spike = self.avgpool1(spike)
            spike = self.conv3(spike)
            spike = self.conv4(spike)
            spike = self.conv5(spike)

            fc_in = self.maxpool2(spike).view(input_x.size(0), -1)

            character = self.fc1000(fc_in)
            out = self.classifier(character)
        return out / self.classifier.threshold / steps
    
    def tst(self, input_x, steps=100, l=1):
        # 注意这里没有reset 
        # 效果跑完再和上面合并，并且out可以测试分开conv和集中conv的效果，上面训练也可以改成集中conv
        self.reset_mem()
        out = 0
        with torch.no_grad():
            for i in range(steps * self.self.repeat):
                # Poisson input spike generation
                eventframe_input = (input_x[:, :, :, :, (i % steps)]).float()
                eventframe_input = self.self.avgtimes * self.maxpool1(eventframe_input)
                spike = self.conv1.snn_forward(eventframe_input)
                spike = self.conv2.snn_forward(spike)
                spike = self.avgpool1.snn_forward(spike)
                spike = self.conv3.snn_forward(spike)
                spike = self.conv4.snn_forward(spike)
                spike = self.conv5.snn_forward(spike)

                fc_in = self.maxpool2(spike).view(input_x.size(0), -1)

                spike = self.fc1000.snn_forward(fc_in)
                out += self.classifier(spike)
            return out / self.classifier.threshold / steps
            
    def updateScale(self, sign):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.updateScale(sign)

    def reset_mem(self):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.reset(0, 0, 0)
        

    def _weight_init(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.in_channels
                variance1 = math.sqrt(2.0 / n)
                m.weight.data.normal_(0, variance1)
                # define threshold
                # m.threshold = self.vth

            elif isinstance(m, nn.Linear):
                size = m.weight.size()
                fan_in = size[1]  # number of columns
                variance2 = math.sqrt(2.0 / fan_in)
                m.weight.data.normal_(0.0, variance2)
                # define threshold
                # m.threshold = self.vth 

class SNN_noaltp(nn.Module):
    def __init__(self, size, repeat, avgtimes, sign, device, vth, leak) -> None:
        super(SNN_noaltp, self).__init__()
        self.repeat = repeat
        self.avgtimes = avgtimes
        self.sign = sign
        self.vth = vth
        self.leak = leak
        H,W = size # list:[H,W]
        para_size_H = H
        para_size_W = W
        # kernel_size = 3
        # stride = 1
        # padding = 1
        device = torch.device(f'cuda:{device}')
        self.maxpool4 = nn.AvgPool2d(kernel_size=3)
        para_size_H = int(para_size_H/3)
        para_size_W = int(para_size_W/3)
        conv11 = nn.Conv2d(in_channels=2, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron11 = Neuron(conv11, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv12 = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron12 = Neuron(conv12, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        avgpool1 = nn.AvgPool2d(kernel_size=2)
        self.neuronp1 = Neuron(avgpool1, 0.75, self.leak, pool=True)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        conv21 = nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron21 = Neuron(conv21, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv22 = nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron22 = Neuron(conv22, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv23 = nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron23 = Neuron(conv23, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.avgpool2 = nn.MaxPool2d(kernel_size=2)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)

        fc0 = nn.Linear(256 * para_size_H * para_size_W, 1024, bias=False)
        # shareF0 = ALTP_D((1024), device=device)
        # self.neuronf0 = Neuron(ResponseFunc(fc0, shareF0), self.vth, self.leak)
        self.neuronf0 = Neuron(fc0, self.vth, self.leak)
        self.classifier = nn.Linear(1024, 10, bias=False)
        self.classifier.threshold = self.vth
        
        self._weight_init()

    def forward(self, input_x, steps=100, l=1):
        self.reset_mem()
        total_input = 0
        with torch.no_grad():
            for i in range(steps * self.repeat):
                # Poisson input spike generation
                eventframe_input = input_x[:, :, :, :, i % steps].float()
                eventframe_input = self.avgtimes * self.maxpool4(eventframe_input)
                total_input += eventframe_input
                spike = self.neuron11.snn_forward(eventframe_input)
                spike = self.neuron12.snn_forward(spike)
                spike = self.neuronp1.snn_forward(spike)
                spike = self.neuron21.snn_forward(spike)
                spike = self.neuron22.snn_forward(spike)
                spike = self.neuron23.snn_forward(spike)

                fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

                spike = self.neuronf0.snn_forward(fc_in)
            self.updateScale(self.sign)
        with torch.enable_grad():
            spike = self.neuron11(total_input)
            spike = self.neuron12(spike)
            spike = self.neuronp1(spike)
            spike = self.neuron21(spike)
            spike = self.neuron22(spike)
            spike = self.neuron23(spike)

            fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

            character = self.neuronf0(fc_in)
            out = self.classifier(character)
        return out / self.classifier.threshold / steps
    
    def tst(self, input_x, steps=100, l=1):
        # 注意这里没有reset 
        # 效果跑完再和上面合并，并且out可以测试分开conv和集中conv的效果，上面训练也可以改成集中conv
        self.reset_mem()
        out = 0
        with torch.no_grad():
            for i in range(steps * self.repeat):
                # Poisson input spike generation
                eventframe_input = (input_x[:, :, :, :, (i % steps)]).float()
                eventframe_input = self.avgtimes * self.maxpool4(eventframe_input)
                spike = self.neuron11.snn_forward(eventframe_input)
                spike = self.neuron12.snn_forward(spike)
                spike = self.neuronp1.snn_forward(spike)
                spike = self.neuron21.snn_forward(spike)
                spike = self.neuron22.snn_forward(spike)
                spike = self.neuron23.snn_forward(spike)

                fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

                spike = self.neuronf0.snn_forward(fc_in)
                out += self.classifier(spike)
            return out / self.classifier.threshold / steps
        

    def updateScale(self, sign):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.updateScale(sign)

    def reset_mem(self):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.reset(0, 0, 0)
        

    def _weight_init(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.in_channels
                variance1 = math.sqrt(2.0 / n)
                m.weight.data.normal_(0, variance1)
                # define threshold
                # m.threshold = self.vth

            elif isinstance(m, nn.Linear):
                size = m.weight.size()
                fan_in = size[1]  # number of columns
                variance2 = math.sqrt(2.0 / fan_in)
                m.weight.data.normal_(0.0, variance2)
                # define threshold
                # m.threshold = self.vth 

class SNN_VGG11_noaltp(nn.Module):
    def __init__(self, size, repeat, avgtimes, sign, device, vth, leak) -> None:
        super(SNN_VGG11_noaltp, self).__init__()
        self.repeat = repeat
        self.avgtimes = avgtimes
        self.sign = sign
        self.vth = vth
        self.leak = leak
        H,W = size # list:[H,W]
        para_size_H = H
        para_size_W = W
        device = torch.device(f'cuda:{device}')
        # input tansform
        self.maxpool4 = nn.AvgPool2d(kernel_size=3)
        para_size_H = int(para_size_H/3)
        para_size_W = int(para_size_W/3)
        # block1
        conv11 = nn.Conv2d(in_channels=2, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron11 = Neuron(conv11, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        avgpool1 = nn.AvgPool2d(kernel_size=2)
        self.neuronp1 = Neuron(avgpool1, 0.75, self.leak, pool=True)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        # block2
        conv21 = nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron21 = Neuron(conv21, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        avgpool2 = nn.AvgPool2d(kernel_size=2)
        self.neuronp1 = Neuron(avgpool2, 0.75, self.leak, pool=True)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        # block3
        conv31 = nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron31 = Neuron(conv31, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv32 = nn.Conv2d(in_channels=256, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron32 = Neuron(conv32, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        avgpool3 = nn.AvgPool2d(kernel_size=2)
        self.neuronp3 = Neuron(avgpool3, 0.75, self.leak, pool=True)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        # block4
        conv41 = nn.Conv2d(in_channels=256, out_channels=512, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron41 = Neuron(conv41, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv42 = nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron42 = Neuron(conv42, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.avgpool4 = nn.MaxPool2d(kernel_size=2)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        # block5
        conv51 = nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron51 = Neuron(conv51, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        conv52 = nn.Conv2d(in_channels=512, out_channels=512, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron52 = Neuron(conv52, self.vth, self.leak)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.avgpool5 = nn.MaxPool2d(kernel_size=2)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        #fully connected
        fc0 = nn.Linear(512 * para_size_H * para_size_W, 4096, bias=False)
        # shareF0 = ALTP_D((1024), device=device)
        # self.neuronf0 = Neuron(ResponseFunc(fc0, shareF0), self.vth, self.leak)
        self.neuronf0 = Neuron(fc0, self.vth, self.leak)
        self.dropout1 = nn.Dropout(p=0.5, inplace=False)
        fc1 = nn.Linear(4096, 4096, bias=False)
        self.neuronf1 = Neuron(fc1, self.vth, self.leak)
        self.dropout1 = nn.Dropout(p=0.5, inplace=False)
        self.classifier = nn.Linear(4096, 4096, bias=False)



        self.classifier.threshold = self.vth
        
        self._weight_init()

    def forward(self, input_x, steps, l=1):
        self.reset_mem()
        total_input = 0
        with torch.no_grad():
            for i in range(steps * self.repeat):
                # Poisson input spike generation
                eventframe_input = input_x[:, :, :, :, i % steps].float()
                eventframe_input = self.avgtimes * self.maxpool4(eventframe_input)
                total_input += eventframe_input
                spike = self.neuron11.snn_forward(eventframe_input)
                spike = self.neuron12.snn_forward(spike)
                spike = self.neuronp1.snn_forward(spike)
                spike = self.neuron21.snn_forward(spike)
                spike = self.neuron22.snn_forward(spike)
                spike = self.neuron23.snn_forward(spike)

                fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

                spike = self.neuronf0.snn_forward(fc_in)
            self.updateScale(self.sign)
        with torch.enable_grad():
            spike = self.neuron11(total_input)
            spike = self.neuron12(spike)
            spike = self.neuronp1(spike)
            spike = self.neuron21(spike)
            spike = self.neuron22(spike)
            spike = self.neuron23(spike)

            fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

            character = self.neuronf0(fc_in)
            out = self.classifier(character)
        return out / self.classifier.threshold / steps
    
    def tst(self, input_x, steps=100, l=1):
        # 注意这里没有reset 
        # 效果跑完再和上面合并，并且out可以测试分开conv和集中conv的效果，上面训练也可以改成集中conv
        self.reset_mem()
        out = 0
        with torch.no_grad():
            for i in range(steps * self.repeat):
                # Poisson input spike generation
                eventframe_input = (input_x[:, :, :, :, (i % steps)]).float()
                eventframe_input = self.avgtimes * self.maxpool4(eventframe_input)
                spike = self.neuron11.snn_forward(eventframe_input)
                spike = self.neuron12.snn_forward(spike)
                spike = self.neuronp1.snn_forward(spike)
                spike = self.neuron21.snn_forward(spike)
                spike = self.neuron22.snn_forward(spike)
                spike = self.neuron23.snn_forward(spike)

                fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

                spike = self.neuronf0.snn_forward(fc_in)
                out += self.classifier(spike)
            return out / self.classifier.threshold / steps
        

    def updateScale(self, sign):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.updateScale(sign)

    def reset_mem(self):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.reset(0, 0, 0)
        

    def _weight_init(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.in_channels
                variance1 = math.sqrt(2.0 / n)
                m.weight.data.normal_(0, variance1)
                # define threshold
                # m.threshold = self.vth

            elif isinstance(m, nn.Linear):
                size = m.weight.size()
                fan_in = size[1]  # number of columns
                variance2 = math.sqrt(2.0 / fan_in)
                m.weight.data.normal_(0.0, variance2)
                # define threshold
                # m.threshold = self.vth 

class VGG7_bn(nn.Module):
    def __init__(self, size, in_channel=2,num_classes=10) -> None:
        nn.Module.__init__(self)
        H,W = size # list:[H,W]
        para_size_H = H
        para_size_W = W
        # self.maxpool_init = nn.AvgPool2d(kernel_size=3)
        # para_size_H = int(para_size_H/3)
        # para_size_W = int(para_size_W/3)
        self.conv11 = nn.Conv2d(in_channels=in_channel, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn11 = nn.BatchNorm2d(64)
        self.act11 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.conv12 = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn12 = nn.BatchNorm2d(64)
        self.act12 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.maxpool1 = nn.MaxPool2d(kernel_size=2)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        self.conv21 = nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn21 = nn.BatchNorm2d(128)
        self.act21 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.conv22 = nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn22 = nn.BatchNorm2d(128)
        self.act22 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.conv23 = nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn23 = nn.BatchNorm2d(256)
        self.act23 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.maxpool2 = nn.MaxPool2d(kernel_size=2)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)

        self.fc0 = nn.Linear(256 * para_size_H * para_size_W, 1024, bias=False)
        self.actf0 = nn.ReLU(inplace=True)
        self.fc1 = nn.Linear(1024, num_classes, bias=False)
        self.actf1 = nn.ReLU(inplace=True)
    
    def forward(self,input):
        # input transform
        # input_init = self.maxpool_init(input)
        # block1
        x = self.act11(self.bn11(self.conv11(input)))
        x = self.act12(self.bn12(self.conv12(x)))
        x = self.maxpool1(x)
        # block2
        x = self.act21(self.bn21(self.conv21(x)))
        x = self.act22(self.bn22(self.conv22(x)))
        x = self.act23(self.bn23(self.conv23(x)))
        x = self.maxpool2(x)
        # fully connected
        x = x.view(x.size(0), -1)
        x = self.actf0(self.fc0(x))
        output = self.actf1(self.fc1(x))
        # output
        return output

class VGG7(nn.Module):
    def __init__(self, size, in_channel=2, num_classes=10) -> None:
        nn.Module.__init__(self)
        H,W = size # list:[H,W]
        para_size_H = H
        para_size_W = W
        # self.maxpool_init = nn.AvgPool2d(kernel_size=3)
        # para_size_H = int(para_size_H/3)
        # para_size_W = int(para_size_W/3)
        self.conv11 = nn.Conv2d(in_channels=in_channel, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.act11 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.conv12 = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        self.act12 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.maxpool1 = nn.MaxPool2d(kernel_size=2)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)
        self.conv21 = nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.act21 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.conv22 = nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.act22 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.conv23 = nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False)
        self.act23 = nn.ReLU(inplace=True)
        para_size_H = int((para_size_H - 3 + 2 * 1 )/1) + 1
        para_size_W = int((para_size_W - 3 + 2 * 1 )/1) + 1
        self.maxpool2 = nn.MaxPool2d(kernel_size=2)
        para_size_H = int(para_size_H/2)
        para_size_W = int(para_size_W/2)

        self.fc0 = nn.Linear(256 * para_size_H * para_size_W, 1024, bias=False)
        self.actf0 = nn.ReLU(inplace=True)
        self.fc1 = nn.Linear(1024, num_classes, bias=False)
        self.actf1 = nn.ReLU(inplace=True)
    
    def forward(self,input):
        # input transform
        # input_init = self.maxpool_init(input)
        # block1
        x = self.act11(self.conv11(input))
        x = self.act12(self.conv12(x))
        x = self.maxpool1(x)
        # block2
        x = self.act21(self.conv21(x))
        x = self.act22(self.conv22(x))
        x = self.act23(self.conv23(x))
        x = self.maxpool2(x)
        # fully connected
        x = x.view(x.size(0), -1)
        x = self.actf0(self.fc0(x))
        output = self.actf1(self.fc1(x))
        # output
        return output

 


from collections import OrderedDict
from re import L
from typing import Any, List
from torchvision.models.resnet import resnet18, resnet34
import torch.nn as nn
import torch

from SNN.network.node import Neuron

cfgs = {'resnet9': [
        ('backbone', [64, 'M']),
        ('block', [64, 128]),
        ('block', [128, 256]),
        ('block', [256, 512]),
    ]}

def make_layers(cfg) -> nn.Sequential:
    layers: List[nn.Module] = []
    in_channels = 3
    for m in cfg:
        if m[0] is 'backbone':
            for v in m[1]:
                if v == 'M':
                    layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
                else:
                    conv2d = nn.Conv2d(in_channels, v, kernel_size=3, stride=1, padding=1, bias=False)
                    layers += [conv2d, nn.BatchNorm2d(v), nn.ReLU(inplace=True)]
                    in_channels = v
        elif m[0] is 'block':
            layers += [cnnBasicBlock(m[1][0], m[1][1])]
    return nn.Sequential(*layers)

class cnnBasicBlock(nn.Module):
    def __init__(
        self,
        inplanes: int,
        outplanes: int,
    ) -> None:
        super(cnnBasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, outplanes, kernel_size=3, stride=2, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(outplanes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(outplanes, outplanes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(outplanes)
        self.downsample = nn.Conv2d(inplanes, outplanes, kernel_size=1, stride=2, padding=0, bias=False)
        self.bndown = nn.BatchNorm2d(outplanes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        identity = self.downsample(x)
        identity = self.bndown(identity)

        out += identity
        out = self.relu(out)

        return out

class cnnSmallResNet(nn.Module):
    def __init__(
        self,
        features: nn.Module,
        num_classes: int = 10,
    ) -> None:
        super(cnnSmallResNet, self).__init__()
        self.features = features
        self.avgpool = nn.AdaptiveAvgPool2d((7, 7))
        self.classifier = nn.Sequential(OrderedDict([
          ('fc1', nn.Linear(512*7*7, 1024)),
          ('fc2', nn.Linear(1024, num_classes)),
            ])
        )

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.features(x)
        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x


def resnet9(pretrained: bool = False, progress: bool = True, **kwargs: Any) -> cnnSmallResNet:
    return cnnSmallResNet(make_layers(cfgs['resnet9']), **kwargs)




class ResponseFunc_skip(nn.Module):
    def __init__(self, cnn, skip) -> None:
        super(ResponseFunc_skip, self).__init__()
        self.cnn = cnn
        self.skip = skip
    def forward(self, x):
        return self.cnn(x[0]) + self.skip(x[1])

class snnBlock(nn.Module):
    def __init__(self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        vth:int=1,
        leak:int=1) -> None:
        super(snnBlock, self).__init__()
        # stride should be 1, x, x? x, x, 1
        conv1 = nn.Conv2d(in_channels=inplanes, out_channels=planes, 
            kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron1 = Neuron(conv1, vth, leak)
        conv2 = nn.Conv2d(in_channels=planes, out_channels=planes, 
            kernel_size=3, stride=stride, padding=1, bias=False)
        conv3 = nn.Conv2d(in_channels=inplanes, out_channels=planes, 
            kernel_size=1, stride=stride, padding=0, bias=False)
        self.neuron2 = Neuron(ResponseFunc_skip(conv2, conv3), vth, leak)

    def forward(self, x):
        m = self.neuron1(x)
        return self.neuron2((m, x))
    
    def snn_forward(self, x):
        m = self.neuron1.snn_forward(x)
        return self.neuron2.snn_forward((m, x))

class snnResnet(nn.Module):
    def __init__(self,
        layers: List[int],
        in_channel = 2,
        num_classes:int=10,
        HW:tuple=(42, 42),

        vth:int=1,
        leak:int=1) -> None:

        super(snnResnet, self).__init__()

        conv11 = nn.Conv2d(in_channels=in_channel, out_channels=64, 
            kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron1 = Neuron(conv11, vth, leak)
        avgpool1 = nn.AvgPool2d(kernel_size=2, stride=2)
        self.neuronp1 = Neuron(avgpool1, 0.75, leak, pool=True)
        self.layer1 = self._make_layers(layers[0], 64, 128, stride=1, vth=vth, leak=leak)
        self.layer2 = self._make_layers(layers[1], 128, 256, stride=2, vth=vth, leak=leak)
        self.layer3 = self._make_layers(layers[2], 256, 512, stride=2, vth=vth, leak=leak)
        H, W = int(HW[0]/2), int(HW[1]/2)
        for _ in range(layers[1]+layers[2]):
            H, W = int((H+1)/2), int((W+1)/2)
        fc1 = nn.Linear(512 * H*W, 1024, bias=False)
        self.neuronf1 = Neuron(fc1, vth, leak)
        self.fc2 = nn.Linear(1024, num_classes, bias=False)
        self.vth = vth

    def snn_forward(self, x):
        x = self.neuron1.snn_forward(x)
        x = self.neuronp1.snn_forward(x)
        for m in self.layer1:
            x = m.snn_forward(x)
        for m in self.layer2:
            x = m.snn_forward(x)
        for m in self.layer3:
            x = m.snn_forward(x)
        x = torch.flatten(x, 1)
        x = self.neuronf1.snn_forward(x)
        return x
    
    def forward(self, x):
        x = self.neuron1(x)
        x = self.neuronp1(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = torch.flatten(x, 1)
        x = self.neuronf1(x)
        x = self.fc2(x)
        return x / self.vth
    
    def updateScale(self, sign):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.updateScale(sign)

    def reset_mem(self):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.reset(0, 0, 0)

    def _make_layers(self, num, in_channel, out_channel, stride, vth, leak):
        layers = []
        layers.append(snnBlock(in_channel, out_channel, stride, vth, leak))
        for _ in range(1, num):
            layers.append(snnBlock(out_channel, out_channel, stride, vth, leak))
        return nn.Sequential(*layers)

def snnresnet9(vth, leak, num_classes, CHW)->snnResnet:
    return snnResnet([1, 1, 1], CHW[0], num_classes, (CHW[1], CHW[2]),vth, leak)

def snnresnet18(vth, leak, num_classes, CHW)->snnResnet:
    return snnResnet([3, 2, 2], CHW[0], num_classes, (CHW[1], CHW[2]),vth, leak)



import numpy as np
import torch
import random
import torch.backends.cudnn as cudnn
import os

def setRandom(seed, Faster=True):
    if seed is None:
        seed = random.randint(1, 10000)
    print('seed:', seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    
    if Faster:
        cudnn.benchmark = True
    else:
        cudnn.benchmark = False
        cudnn.deterministic = True

def printArgs(args):
    print('\n'+'='*15+'settings'+'='*15)
    print('pid:', os.getpid())  
    for k in list(vars(args).keys()):
        print('%s: %s' % (k, vars(args)[k]))
    print('='*15+'settings'+'='*15+'\n')

class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


import sys
sys.path.append("/home/shb/SpikingNeuralNetwork/") 

from SNN.datasets import CIFAR10DVS
from SNN.utils.transformer import EventReorder
import argparse
import os
import shutil
import time
import math
import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
from torch.autograd import Variable
import torch.optim
import torch.utils.data
import torchvision
import torchvision.transforms as transforms
import torchvision.models as models
import copy
import numpy as np 
import random

parser = argparse.ArgumentParser(description='PyTorch ALTP')
parser.add_argument('--dataset', default='MNIST', type=str, help='dataset = [MNIST]')
parser.add_argument('--start-epoch', default=0, type=int, metavar='N',
                    help='manual epoch number (useful on restarts)')
parser.add_argument('-b', '--batch-size', default=40, type=int,
                    metavar='N', help='mini-batch size (default: 40)')
parser.add_argument('--weight-decay', '--wd', default=5e-4, type=float,
                    metavar='W', help='weight decay (default: 1e-4)')
parser.add_argument('--resume', default='', type=str, metavar='PATH',
                    help='path to latest checkpoint (default: none)')
parser.add_argument('-load', default='', type=str, metavar='PATH',
                    help='path to training mask (default: none)')
parser.add_argument('-e', '--evaluate', dest='evaluate', action='store_true',
                    help='evaluate model on validation set')
parser.add_argument('--lr', '--learning-rate', default=0.0005, type=float,
                    metavar='LR', help='initial learning rate')
parser.add_argument('-j', '--workers', default=1, type=int, metavar='N',                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                
                    help='number of data loading workers (default: 4)')
parser.add_argument('--epochs', default=120, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('--steps', default=100, type=int, metavar='N',
                    help='number of time steps to run')
parser.add_argument('--vth', default=2, type=float, metavar='vth',
                    help='threshold')
parser.add_argument('--leak', default=1, type=float, metavar='leak',
                    help='leaky parameter')
parser.add_argument('--repeat', default=3, type=int,
                    metavar='N', help='repeat (default: 1)')
parser.add_argument('--count', default=1, type=int,
                    help='count (default: 1)')  # 时间窗口内的脉冲是计数还是百分比（per)  处在后per为0
parser.add_argument('--per', default=0.25, type=float,
                    metavar='per', help='per (default: 0.25)')
parser.add_argument('--avgtimes', default=4, type=float,
                    metavar='avgtimes', help='avgtimes (default: 9)')

parser.add_argument('--hz', default=5, type=int, metavar='hz',
                    help='scale update hz')
parser.add_argument('--device', default=0, type=int, metavar='device',
                    help='device id for gpu')
parser.add_argument('--gpus', default='0', type=str,
                    help='id(s) for CUDA_VISIBLE_DEVICES')
parser.add_argument('--change', type=int, nargs='+', default=[30, 60, 80, 100],
                    help='Decrease learning rate at these epochs.')
parser.add_argument('--lr_gama', type=int, nargs='+', default=[0.2, 0.2, 0.2, 0.2],
                    help='lr change gama times at each point.')
parser.add_argument('--seed', type=int, help='manual seed')
parser.add_argument('--sign', type=int, help='scale sign')
parser.add_argument('--data_path', default='/home/shb/datasets/CIFAR10DVS', type=str,
                    help='data_set_path')        

def SNNpadding(img, border=None, fill=0):
    # CHWT
    if isinstance(border, tuple):
        if len(border) == 2:
            left, top = right, bottom = border
        elif len(border) == 4:
            left, top, right, bottom = border
    else:
        left = top = right = bottom = border
    return np.pad(img, ((0, 0), (top, bottom), (left, right), (0, 0)), mode='constant', constant_values=fill)

class RandomCrop(object):
    """Crop the given DVS at a random location.
    input ndarray
    """

    def __init__(self, size, padding=None, pad_if_needed=False, fill=0, padding_mode='constant'):
        if isinstance(size, tuple):
            self.size = size
        else:
            self.size = (int(size), int(size))
        self.padding = padding
        self.pad_if_needed = pad_if_needed
        self.fill = fill
        self.padding_mode = padding_mode

    @staticmethod
    def get_params(img, output_size):
        """Get parameters for ``crop`` for a random crop.

        Args:
            img : Image to be cropped CHWT.
            output_size (tuple): Expected output size of the crop.

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        h, w = img.shape[1], img.shape[2]
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, img):
        """
        Args:
            img (ndarray): CHWT to be cropped.

        Returns:
            img (ndarray): Cropped image.
        """
        if self.padding is not None:
            img = SNNpadding(img, self.padding, self.fill)

        i, j, h, w = self.get_params(img, self.size)

        return img[:, i:i+h, j:j+w, :].copy()

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, padding={1})'.format(self.size, self.padding)

class RandomTranspose(object):
    """Horizontally flip the given ndarray randomly with a given probability.

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img):
        """
        Args:
            img (ndarray): CHWT to be transpose.

        Returns:
            img (ndarray): transposed image.
        """
        if random.random() < self.p:
            return img.swapaxes(1, 2).copy()
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)

class RandomHorizontalFlip(object):
    """Horizontally flip the given ndarray randomly with a given probability.

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img):
        """
        Args:
            img (ndarray): Image to be flipped.

        Returns:
            ndarray: Randomly flipped image.
        """
        if random.random() < self.p:
            return np.flip(img, 2).copy()
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


args = parser.parse_args()

os.environ["CUDA_VISIBLE_DEVICES"] = args.gpus

def setRandom():
    if args.seed is None:
        args.seed = random.randint(1, 10000)
    print('seed:', args.seed)
    torch.manual_seed(args.seed)
    torch.cuda.manual_seed(args.seed)
    torch.cuda.manual_seed_all(args.seed)
    np.random.seed(args.seed)
    random.seed(args.seed)
    # cudnn.benchmark = False
    # cudnn.deterministic = True
    cudnn.benchmark = True


def main():
    device = torch.device(f'cuda:{args.device}')
    torch.cuda.set_device(device)
    seed1 = 44
    seed2 = 85
    seed3 = 63
    torch.manual_seed(seed1)
    torch.cuda.manual_seed(seed2)
    torch.cuda.manual_seed_all(seed3)
    np.random.seed(seed1)
    random.seed(seed2)
    cudnn.benchmark = False
    cudnn.deterministic = True
    
    print('\n'+'='*15+'settings'+'='*15)
    print('pid:', os.getpid())
    print('lr: ', args.lr)
    print('change lr point:', args.change)
    print('change gama', args.lr_gama)
    print('batchsize:',args.batch_size)
    print('epochs:', args.epochs)
    print('steps:', args.steps)
    print('vth:', args.vth)
    print('leak:{}'.format(args.leak))
    print('hz:{}'.format(args.hz))
    print('repeat:{}'.format(args.repeat))
    print('count={}'.format(args.count))
    print('per={}'.format(args.per))
    print('='*15+'settings'+'='*15+'\n')

    best_prec1 = 0
    ep = []
    lRate = []   
    tp1_5_loss_tr = []
    tp1_5_loss_eval = []

    model = SNN(args)

    print(model)
    
    model.to(device)
    
    criterion = torch.nn.MSELoss(reduction='sum')
    criterion_en = torch.nn.CrossEntropyLoss()

    optimizer = torch.optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.weight_decay) #

    # optionally resume from a checkpoint
    if args.resume:
        if os.path.isfile(args.resume):
            print("=> loading checkpoint '{}'".format(args.resume))
            checkpoint = torch.load(args.resume, map_location=device)
            args.start_epoch = checkpoint['epoch']
            best_prec1 = checkpoint['best_prec1']
            model.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            print("=> loaded checkpoint '{}' (epoch {})"
                  .format(args.resume, checkpoint['epoch']))
            args.warmup = 0
        else:
            print("=> no checkpoint found at '{}'".format(args.resume))

    # Dataloader
    time_read1 = time.time()
    normalize = transforms.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.557, 0.549, 0.5534])
    transform_train = transforms.Compose([
        # RandomTranspose(1),
        # RandomCrop(128, padding=16),
        # RandomHorizontalFlip(),
        # EventReorder((0, 2, 1, 3))
        # transforms.ToTensor(),
        # normalize,
    ])
    train_data = CIFAR10DVS(args.data_path, download=False, step=args.steps, subSet="Train",
                transform=transform_train, target_transform=None)
    train_loader = torch.utils.data.DataLoader(train_data,
                                               batch_size=args.batch_size, shuffle=True,
                                               num_workers=args.workers,
                                               pin_memory=True)
    time_read1 = time.time() - time_read1
    print('read data of testing takes %dh %dmin' %
          (time_read1/3600, (time_read1 % 3600)/60))

    time_read1 = time.time()
    transform_test = transforms.Compose([
        # transforms.ToTensor(),
        # normalize,
    ])
    val_data = CIFAR10DVS(args.data_path, step = args.steps, subSet="Test", 
                        transform=transform_test)
    val_loader = torch.utils.data.DataLoader(val_data,  # val_data for testing
                                             batch_size=int(args.batch_size/2), shuffle=False,
                                             num_workers=args.workers,
                                             pin_memory=False)
    time_read1 = time.time() - time_read1
    print('read data of testing takes %dh %dmin' %
          (time_read1/3600, (time_read1 % 3600)/60))
    print('read dataset done')
    if args.evaluate:
        validate(val_loader, model, criterion, criterion_en, time_steps=args.steps, leak=args.leak, device=device)
        return

    for epoch in range(args.start_epoch, args.epochs):
        for name, wmax, wmin, wmean in getALTPw(model):
            print(name, wmax, wmin, wmean)
        if epoch % args.hz == 0 and args.hz < args.epochs:
            args.sign = 1
        else:
            args.sign = 0

        start = time.time()
        args.lr = adjust_learning_rate(optimizer, epoch, args.lr)
        lRate.append(args.lr)
        ep.append(epoch)

        # train for one epoch
        tp1_5_loss_tr.append(train(train_loader, model, criterion, criterion_en, optimizer, epoch, 
                            time_steps=args.steps, leak=args.leak, device=device))

        # evaluate on validation set
        modeltest = model
        tp1_5_loss_eval.append(validate(val_loader, modeltest, criterion, criterion_en, 
                                time_steps=args.steps, leak=args.leak, device=device))
        prec1 = tp1_5_loss_eval[-1][0]

        # remember best prec@1 and save checkpoint
        is_best = prec1 > best_prec1
        best_prec1 = max(prec1, best_prec1)
        save_checkpoint({
            'epoch': epoch + 1,
            'state_dict': model.state_dict(),
            'best_prec1': best_prec1,
            'optimizer': optimizer.state_dict(),
        }, is_best, epoch, prec1)
        time_use = time.time() - start
        print('time used this epoch: %d h%dmin%ds' %(time_use//3600,(time_use%3600)//60,time_use%60))

        if args.sign:
            print('\n'+'='*15+'scale'+'='*15)            
            for name, scale in getScales(model):
                print(name, scale)
            print('='*15+'scale'+'='*15+'\n')


    for k in range(0, args.epochs - args.start_epoch):
        print('Epoch: [{0}/{1}]\t'
              'LR:{2}\t'
              'Prec@1 {top1:.3f} \t'
              'Prec@5 {top5:.3f} \t'
              'En_Loss_Eval {losses_en_eval: .4f} \t'
              'Prec@1_tr {top1_tr:.3f} \t'
              'Prec@5_tr {top5_tr:.3f} \t'
              'En_Loss_train {losses_en: .4f}'.format(
            ep[k], args.epochs, lRate[k], top1=tp1_5_loss_eval[k][0], top5=tp1_5_loss_eval[k][1], losses_en_eval=tp1_5_loss_eval[k][2], 
            top1_tr=tp1_5_loss_tr[k][0], top5_tr=tp1_5_loss_tr[k][1], losses_en=tp1_5_loss_tr[k][2]))
    print('best_acc={}'.format(best_prec1))



def grad_cal(scale, IF_in):
    out = scale * IF_in.gt(0).type(torch.cuda.FloatTensor)
    return out

def ave(output, input):
    c = input >= output
    if input[c].sum() < 1e-3:
        return 1
    return output[c].sum()/input[c].sum()


def train(train_loader, model, criterion, criterion_en, optimizer, epoch, time_steps, leak, device):
    # switch to train mode
    model.train()

    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    top5 = AverageMeter()

    top1_tr = AverageMeter()
    top5_tr = AverageMeter()
    losses_en = AverageMeter()


    end = time.time()
    start_end = end 
    for i, (input, target) in enumerate(train_loader):
        # if args.warmup != 0 and args.warm_iter < args.warmup:
        #     adjust_warmup_lr(optimizer, args.warm_iter, args.warmup)
        #     args.warm_iter += 1
        # measure data loading time
        data_time.update(time.time() - end)
        input, target = input.to(device), target.to(device)
        # input, targets_a, targets_b, lam = mixup_data(input, target, 1.)
        label = target.clone()
        # labels_b = targets_b.clone()

        optimizer.zero_grad()  # Clear gradients w.r.t. parameters

        output = model(input, steps=time_steps, l=leak)
        
        targetN = output.data.clone().zero_().to(device)
        targetN.scatter_(1, target.unsqueeze(1), 1)
        targetN = Variable(targetN.type(torch.cuda.FloatTensor))

        # targetN_b = output.data.clone().zero_().to(device)
        # targetN_b.scatter_(1, targets_b.unsqueeze(1), 1)
        # targetN_b = Variable(targetN_b.type(torch.cuda.FloatTensor))

        # loss = mixup_criterion(criterion, output.cpu(), targetN_a.cpu(), targetN_b.cpu(), lam)
        # loss_en = mixup_criterion(criterion_en, output.cpu(), labels_a.cpu(), labels_b.cpu(), lam)
        loss = criterion(output.cpu(), targetN.cpu())
        loss_en = criterion_en(output.cpu(), label.cpu())

        # measure accuracy and record loss
        prec1, prec5 = accuracy(output.data, target, topk=(1, 5))
        losses.update(loss.item(), input.size(0))
        top1.update(prec1.item(), input.size(0))
        top5.update(prec5.item(), input.size(0))

        prec1_tr, prec5_tr = accuracy(output.data, target, topk=(1, 5))
        losses_en.update(loss_en.item(), input.size(0))
        top1_tr.update(prec1_tr.item(), input.size(0))
        top5_tr.update(prec5_tr.item(), input.size(0))

        # compute gradient and do SGD step
        loss.backward(retain_graph=False)
        optimizer.step()


        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()
    
    
    print('Epoch: [{0}] Prec@1 {top1_tr.avg:.3f} Prec@5 {top5_tr.avg:.3f} Entropy_Loss {loss_en.avg:.4f}'
          .format(epoch, top1_tr=top1_tr, top5_tr=top5_tr, loss_en=losses_en))
    time_use = end - start_end
    print('train time: %d h%dmin%ds' %(time_use//3600,(time_use%3600)//60,time_use%60))

    return top1_tr.avg, top5_tr.avg, losses_en.avg

def validate(val_loader, model, criterion, criterion_en, time_steps, leak, device):
    batch_time = AverageMeter()
    data_time = AverageMeter()
    losses = AverageMeter()
    top1 = AverageMeter()
    top5 = AverageMeter()
    losses_en_eval = AverageMeter()

    # switch to evaluate mode
    model.eval()

    end = time.time()
    with torch.no_grad():
        for i, (input, target) in enumerate(val_loader):
	        # measure data loading time
            data_time.update(time.time() - end)
            input_var = input.to(device)
            labels = Variable(target.to(device))
            target = target.to(device)
            output = model.tst(input_var, steps=time_steps, l=leak)

            targetN = output.data.clone().zero_().to(device)
            targetN.scatter_(1, target.unsqueeze(1), 1)
            targetN = Variable(targetN.type(torch.cuda.FloatTensor))
            loss = criterion(output.cpu(), targetN.cpu())
            loss_en = criterion_en(output.cpu(), labels.cpu())
	        # measure accuracy and record loss
            prec1, prec5 = accuracy(output.data, target, topk=(1, 5))
            losses.update(loss.item(), input.size(0))
            top1.update(prec1.item(), input.size(0))
            top5.update(prec5.item(), input.size(0))
            losses_en_eval.update(loss_en.item(), input.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

    print('Test: Prec@1 {top1.avg:.3f} Prec@5 {top5.avg:.3f} Entropy_Loss {losses_en_eval.avg:.4f}'
          .format(top1=top1, top5=top5, losses_en_eval=losses_en_eval))

    return top1.avg, top5.avg, losses_en_eval.avg


def save_checkpoint(state, is_best, epoch, prec1, filename='checkpoint_latest.pth.tar'):
    path = "./ckp/"
    if not os.path.exists(path):
        os.mkdir(path)
    torch.save(state,path + filename)
    if is_best:
        shutil.copyfile(path + filename, path + "best_" + str(epoch) + '_' + str(prec1) + '_' + filename)


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def adjust_learning_rate(optimizer, epoch, lr):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    if epoch in args.change:
        id = args.change.index(epoch)
        lr *= args.lr_gama[id]
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr
    return lr

def adjust_warmup_lr(optimizer, citer, warmup):
    print('warm up:', args.lr * (citer + 1.) / warmup)
    for param_group in optimizer.param_groups:
        param_group['lr'] = args.lr * (citer + 1.) / warmup

def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].reshape(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res

class Neuron(nn.Module):
    def __init__(self, response_func, threshold, leak, _scale=1, pool=False, out=False, name="neuron") -> None:
        super(Neuron, self).__init__()
        self.response_func = response_func
        self.threshold = threshold
        self.leak = leak
        self.name = name
        self._scale = torch.nn.parameter.Parameter(torch.tensor(_scale), requires_grad=False)
        self.pool = pool
        self.out = out
    
    def reset(self, soma_in, mem, soma_out):
        # torch.autograd.Variable(torch.zeros((batch_size, *self.in_shape), device=device), requires_grad=False)  # will not add in optimizer
        self.soma_in = soma_in
        self.mem = mem
        self.soma_out = soma_out

    def snn_forward(self, response_func_input):
        soma_in = self.response_func(response_func_input)
        self.soma_in += soma_in
        self.mem += soma_in
        # 后面再修改threshold， gt成一个函数
        ex_membrane = nn.functional.threshold(self.mem, self.threshold, 0)
        self.mem = self.mem - ex_membrane # hard reset
        # generate spike
        soma_out = ex_membrane.gt(0).type(torch.cuda.FloatTensor)
        if not self.pool:
            self.mem = self.leak * self.mem
        self.soma_out += soma_out
        return soma_out
    
    def updateScale(self, sign):
        if sign:
            self._scale.data = 0.6 * ave(self.soma_out, self.soma_in) + 0.4 * self._scale.data
        self.scale = grad_cal(self._scale.data, self.soma_in)

    def forward(self, sum_x):
        sum_y = self.response_func(sum_x)
        out = torch.mul(sum_y, self.scale)
        return out - out.detach() + self.soma_out.detach()
 
class ALTP_D(nn.Module):
    def __init__(self, shape, ones=True, bias: bool = False, device=None) -> None:
        factory_kwargs = {'device': device}
        super(ALTP_D, self).__init__()
        self.shape = shape
        self.ones = ones
        self.weight = torch.nn.parameter.Parameter(torch.empty(shape, **factory_kwargs))
        if bias:
            self.bias = torch.nn.parameter.Parameter(torch.empty(shape, **factory_kwargs))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        if self.ones:
            torch.nn.init.normal_(self.weight, mean=0.2339, std=0.3971) # from s150's ckp
        else:
            torch.nn.init.zeros_(self.weight)
        if self.bias is not None:
            torch.nn.init.zeros_(self.bias)

    def forward(self, input: torch.Tensor):
        re = input.mul(self.weight.relu())
        if self.bias is not None:
            re = re.add(self.bias)
        return re

    def extra_repr(self) -> str:
        return 'shape={}, bias={}'.format(self.shape, self.bias is not None)

class ResponseFunc(nn.Module):
    def __init__(self, separate, share_func) -> None:
        super(ResponseFunc, self).__init__()
        self.separate = separate
        self.share_func = share_func

    def forward(self, x):
        out = self.separate(x)
        return self.share_func(out.detach())+out

def getScales(model: nn.Module):
    re = []
    for name, m in model.named_modules():
        if isinstance(m, Neuron):
            re.append((name, m._scale))
    return re

def getALTPw(model: nn.Module):
    re = []
    for name, m in model.named_modules():
        if isinstance(m, ALTP_D):
            re.append((name, m.weight.max(), m.weight.min(), m.weight.mean()))
    return re

def mixup_data(x, y, alpha=1.0, use_cuda=True):
    '''Returns mixed inputs, pairs of targets, and lambda'''
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    batch_size = x.size()[0]
    if use_cuda:
        index = torch.randperm(batch_size).cuda()
    else:
        index = torch.randperm(batch_size)

    mixed_x = lam * x + (1 - lam) * x[index, :]
    y_a, y_b = y, y[index]
    return mixed_x, y_a, y_b, lam

def mixup_criterion(criterion, pred, y_a, y_b, lam):
    return lam * criterion(pred, y_a) + (1 - lam) * criterion(pred, y_b)


class SNN(nn.Module):
    def __init__(self, args) -> None:
        super(SNN, self).__init__()
        device = torch.device(f'cuda:{args.device}')
        self.maxpool4 = nn.AvgPool2d(kernel_size=3)
        conv11 = nn.Conv2d(in_channels=2, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        # share11 = ALTP((args.batch_size, 64, 32, 32), device)
        self.neuron11 = Neuron(conv11, args.vth, args.leak)
        conv12 = nn.Conv2d(in_channels=64, out_channels=64, kernel_size=3, stride=1, padding=1, bias=False)
        # share12 = ALTP((args.batch_size, 64, 32, 32), device)
        self.neuron12 = Neuron(conv12, args.vth, args.leak)
        avgpool1 = nn.AvgPool2d(kernel_size=2)
        self.neuronp1 = Neuron(avgpool1, 0.75, args.leak, pool=True)
        
        conv21 = nn.Conv2d(in_channels=64, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron21 = Neuron(conv21, args.vth, args.leak)
        conv22 = nn.Conv2d(in_channels=128, out_channels=128, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron22 = Neuron(conv22, args.vth, args.leak)
        conv23 = nn.Conv2d(in_channels=128, out_channels=256, kernel_size=3, stride=1, padding=1, bias=False)
        self.neuron23 = Neuron(conv23, args.vth, args.leak)
        self.avgpool2 = nn.MaxPool2d(kernel_size=2)

        fc0 = nn.Linear(256 * 10 * 10, 1024, bias=False)
        shareF0 = ALTP_D((1024), device=device)
        self.neuronf0 = Neuron(ResponseFunc(fc0, shareF0), args.vth, args.leak)
        self.classifier = nn.Linear(1024, 10, bias=False)
        self.classifier.threshold = args.vth
        
        self._weight_init()

    def forward(self, input_x, steps, l=1):
        self.reset_mem()
        total_input = 0
        with torch.no_grad():
            for i in range(steps * args.repeat):
                # Poisson input spike generation
                eventframe_input = input_x[:, :, :, :, i % steps].float()
                eventframe_input = args.avgtimes * self.maxpool4(eventframe_input)
                total_input += eventframe_input
                spike = self.neuron11.snn_forward(eventframe_input)
                spike = self.neuron12.snn_forward(spike)
                spike = self.neuronp1.snn_forward(spike)
                spike = self.neuron21.snn_forward(spike)
                spike = self.neuron22.snn_forward(spike)
                spike = self.neuron23.snn_forward(spike)

                fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

                spike = self.neuronf0.snn_forward(fc_in)
            self.updateScale(args.sign)
        with torch.enable_grad():
            spike = self.neuron11(total_input)
            spike = self.neuron12(spike)
            spike = self.neuronp1(spike)
            spike = self.neuron21(spike)
            spike = self.neuron22(spike)
            spike = self.neuron23(spike)

            fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

            character = self.neuronf0(fc_in)
            out = self.classifier(character)
        return out / self.classifier.threshold / steps
    
    def tst(self, input_x, steps=100, l=1):
        # 注意这里没有reset 
        # 效果跑完再和上面合并，并且out可以测试分开conv和集中conv的效果，上面训练也可以改成集中conv
        self.reset_mem()
        out = 0
        with torch.no_grad():
            for i in range(steps * args.repeat):
                # Poisson input spike generation
                eventframe_input = (input_x[:, :, :, :, (i % steps)]).float()
                eventframe_input = args.avgtimes * self.maxpool4(eventframe_input)
                spike = self.neuron11.snn_forward(eventframe_input)
                spike = self.neuron12.snn_forward(spike)
                spike = self.neuronp1.snn_forward(spike)
                spike = self.neuron21.snn_forward(spike)
                spike = self.neuron22.snn_forward(spike)
                spike = self.neuron23.snn_forward(spike)

                fc_in = self.avgpool2(spike).view(input_x.size(0), -1)

                spike = self.neuronf0.snn_forward(fc_in)
                out += self.classifier(spike)
            return out / self.classifier.threshold / steps
        

    def updateScale(self, sign):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.updateScale(sign)

    def reset_mem(self):
        for m in self.modules():
            if isinstance(m, Neuron):
                m.reset(0, 0, 0)
        

    def _weight_init(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.in_channels
                variance1 = math.sqrt(2.0 / n)
                m.weight.data.normal_(0, variance1)
                # define threshold
                # m.threshold = args.vth

            elif isinstance(m, nn.Linear):
                size = m.weight.size()
                fan_in = size[1]  # number of columns
                variance2 = math.sqrt(2.0 / fan_in)
                m.weight.data.normal_(0.0, variance2)
                # define threshold
                # m.threshold = args.vth 


if __name__ == '__main__':
    main()



import numpy as np
import random
from . import event as ev

__all__ = ["EventReorder", "EventReverse", "EventDrift"]


'''combined strategy'''
class EventReorder:
    def __init__(self, p='p', x='x', y='y', t='t') -> None:
        permit = ['p', 'x', 'y', 't']
        assert p in permit and x in permit and y in permit and t in permit, f"args can only in {permit}."
        self.p, self.x, self.y, self.t = p, x, y, t

    def __call__(self, evdata:ev.event):
        return ev.event(evdata[self.p], evdata[self.x], evdata[self.y], evdata[self.t])
    
    def __repr__(self):
        return self.__class__.__name__ + f'\treorder: {self.p}, {self.x}, {self.y}, {self.t}'


'''
reverse space(x, y, xy-yx), temporal(t, p)
random (yes or not) (super range) (random choice)
range 比例的内部实验。
'''

class EventReverse:
    identical = 0
    randomReverse = 1
    xRange = 2
    yRange = 3
    xyRange= 4
    pRange = 5
    xAll = 6 # useful
    yAll = 7 # useless
    pAll = 8
    tALL = 9
    tRange = 10
    def __init__(self, HW, methods, probabilities:list=None, number=1, uniform=False):
        '''
        d are ['t', 'x', 'y', 'p', 'xy']
        events filter are "r-random" "rr-range ratio" "a-all"
        '''
        number = min(len(methods), number)
        self.number = number
        self.p = probabilities
        self.HW = HW
        self.uniform = uniform
        self.method_dict = {EventReverse.identical:['i',(), 0],
            EventReverse.randomReverse:['r',(), 0.3],
            EventReverse.xRange: ['rr', ('x'), 0.5],
            EventReverse.yRange: ['rr', ('y'), 0.5],
            EventReverse.xyRange: ['rr', ('x', 'y'), 0.5],
            EventReverse.pRange: ['rr', ('p'), 0.5],
            EventReverse.xAll: ['a', ('x'), 1],
            EventReverse.yAll: ['a', ('y'), 1],
            EventReverse.tALL: ['a', ('t'), 1],
            EventReverse.pAll: ['a', ('p'), 1],
            EventReverse.tRange: ['rr', ('t'), 0.5]
            }
        if isinstance(methods, list) or isinstance(methods, tuple):
            self.methods = methods
        elif isinstance(methods, dict):
            self.methods = list(methods.keys())
            for m in self.methods:
                self.method_dict[m][2] = methods[m]

    def analyse(self, methods):
        d_events = []
        for m in methods:
            d_events.append(self.method_dict[m])
        return d_events

    def __call__(self, evdata:ev.event):
        method = list(np.random.choice(self.methods, self.number, replace=False, p=self.p))
        d_event = self.analyse(method)
        for elem in d_event:
            if elem[0] == 'i':
                return evdata
                # pass
            elif elem[0] == 'r':
                if self.uniform:
                    elem[2] = np.random.uniform(0, elem[2])
                domains = np.random.choice(('x', 'y', 't', 'p'), np.random.randint(4),replace=False)
                select_events = np.random.choice(evdata.length, int(evdata.length*elem[2]),replace=False)               
                for d in domains:
                    evdata.domainReverse(d, self.HW, select_events)
            elif elem[0] == 'rr':
                domains = elem[1]
                if self.uniform:
                    elem[2] = np.random.uniform(0, elem[2])
                start = np.random.randint(int(evdata.length*(1-elem[2]))+1)
                select_events = slice(start, start+int(evdata.length*elem[2]))  
                for d in domains:
                    evdata.domainReverse(d, self.HW, select_events)
            else: # a
                domains = elem[1]
                select_events = slice(evdata.length)
                for d in domains:
                    evdata.domainReverse(d, self.HW, select_events)
        return evdata

    def __repr__(self):
        return self.__class__.__name__ + f'\ttransformer times:{self.number}\tHW:{self.HW}\
            \nmethods:{self.analyse(self.methods)}\nprobabilities:{self.p}\nuniform:{self.uniform}'

'''
erase + drop: drift(/reverse event/no trigger) out, drift specific, 
drift recurrently / parallelly
drift space:(x, y, x+y), temporal (t)
random (yes or not)(super range) (random choice)
'''    
class EventDrift:
    identical = 0
    randomDrift = 1
    xRange = 2
    yRange = 3
    xyRange= 4
    pRange = 5
    xAll = 6
    yAll = 7
    pAll = 8
    tALL = 9
    xyALL = 10
    # tRange = 11
    # xyFiled = 12
    # current = 11
    # distance = 0.1
    def __init__(self, HW, methods, probabilities=None, number=1, uniform=False):
        number = min(len(methods), number)
        self.number = number
        self.p = probabilities
        self.HW = HW
        self.uniform = uniform
        self.method_dict = {EventDrift.identical:['i',(), 0,0,False],
            EventDrift.randomDrift:['r',(), 0.3, 0.2, False],
            EventDrift.xRange: ['rr', ('x'), 0.5, 0.15, False],
            EventDrift.yRange: ['rr', ('y'), 0.5, 0.15, False],
            EventDrift.xyRange: ['rr', ('x', 'y'), 0.5, 0.15, False],
            EventDrift.pRange: ['rr', ('p'), 0.5, 0.15, False],
            EventDrift.xAll: ['a', ('x'), 1, 0.15, False],
            EventDrift.yAll: ['a', ('y'), 1, 0.15, False],
            EventDrift.tALL: ['a', ('t'), 1, 0.15, False],
            EventDrift.pAll: ['a', ('p'), 1, 0.15, False],
            EventDrift.xyALL: ['a', ('x', 'y'), 1, 0.15, False],
            }
        if isinstance(methods, list) or isinstance(methods, tuple):
            self.methods = methods
        elif isinstance(methods, dict):
            self.methods = list(methods.keys())
            for m in self.methods:
                self.method_dict[m][2:] = methods[m]

    def analyse(self, methods):
        d_events = []
        for m in methods:
            d_events.append(self.method_dict[m])
        return d_events

    def __call__(self, evdata:ev.event):
        method = list(np.random.choice(self.methods, self.number, replace=False, p=self.p))
        d_event = self.analyse(method)
        for elem in d_event:
            if elem[0] == 'i':
                return evdata
                # pass
            elif elem[0] == 'r':
                if self.uniform:
                    elem[2] = np.random.uniform(0, elem[2])
                domains = np.random.choice(('x', 'y', 't', 'p'), np.random.randint(4),replace=False)
                select_events = np.random.choice(evdata.length, int(evdata.length*elem[2]),replace=False)               
                for d in domains:
                    evdata.domainDrift(d, self.HW, select_events, elem[3], elem[4])
            elif elem[0] == 'rr':
                domains = elem[1]
                if self.uniform:
                    elem[2] = np.random.uniform(0, elem[2])
                start = np.random.randint(int(evdata.length*(1-elem[2]))+1)
                select_events = slice(start, start+int(evdata.length*elem[2]))  
                for d in domains:
                    evdata.domainDrift(d, self.HW, select_events, elem[3], elem[4])
            else: # a
                domains = elem[1]
                select_events = slice(evdata.length)
                for d in domains:
                    evdata.domainDrift(d, self.HW, select_events, elem[3], elem[4])
        return evdata

    def __repr__(self):
        return self.__class__.__name__ + f'\ttransformer times:{self.number}\tHW:{self.HW}\
            \nmethods:{self.analyse(self.methods)}\nprobabilities:{self.p}'

class EventDrop:
    def __init__(self, HW):
        self.HW = HW

    def __call__(self, evdata:ev.event):
        evdata.event_drop(self.HW)
        return evdata
    
    def __repr__(self):
        return self.__class__.__name__ + f'\tHW:{self.HW}'

class RandomCrop(object):
    """Crop the given DVS at a random location.
    input ndarray
    """

    def __init__(self, size, padding=None, pad_if_needed=False, fill=0, padding_mode='constant'):
        if isinstance(size, tuple):
            self.size = size
        else:
            self.size = (int(size), int(size))
        self.padding = padding
        self.pad_if_needed = pad_if_needed
        self.fill = fill
        self.padding_mode = padding_mode

    @staticmethod
    def get_params(img, output_size):
        """Get parameters for ``crop`` for a random crop.

        Args:
            img : Image to be cropped CHWT.
            output_size (tuple): Expected output size of the crop.

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        h, w = img.shape[1], img.shape[2]
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def SNNpadding(img, border=None, fill=0):
        # CHWT
        if isinstance(border, tuple):
            if len(border) == 2:
                left, top = right, bottom = border
            elif len(border) == 4:
                left, top, right, bottom = border
        else:
            left = top = right = bottom = border
        return np.pad(img, ((0, 0), (top, bottom), (left, right), (0, 0)), mode='constant', constant_values=fill)

    def __call__(self, img):
        """
        Args:
            img (ndarray): CHWT to be cropped.

        Returns:
            img (ndarray): Cropped image.
        """
        if self.padding is not None:
            img = self.SNNpadding(img, self.padding, self.fill)

        i, j, h, w = self.get_params(img, self.size)

        return img[:, i:i+h, j:j+w, :].copy()

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, padding={1})'.format(self.size, self.padding)

class RandomTranspose(object):
    """Horizontally flip the given ndarray randomly with a given probability.

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img):
        """
        Args:
            img (ndarray): CHWT to be transpose.

        Returns:
            img (ndarray): transposed image.
        """
        if random.random() < self.p:
            return img.swapaxes(1, 2).copy()
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)

class RandomHorizontalFlip(object):
    """Horizontally flip the given ndarray randomly with a given probability.

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """

    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img):
        """
        Args:
            img (ndarray): Image to be flipped.

        Returns:
            ndarray: Randomly flipped image.
        """
        if random.random() < self.p:
            return np.flip(img, 2).copy()
        return img

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


# rely on torch, torchvision

from abc import abstractmethod
from typing import Callable, Dict, List, Optional, Tuple
from torchvision.datasets.utils import download_and_extract_archive, check_integrity, extract_archive ,calculate_md5
from torch.utils.data import Dataset
from .. import event as ev
import os
import numpy as np
import math 

'''
when free, we can provide return classes for functions such as resource. And check function for restructureAndConvert.
'''

def multi_download(func):
    import multiprocessing
    from concurrent.futures import ThreadPoolExecutor
    from functools import wraps
    @wraps(func)
    def decorated(*args, **kwargs):
        pass
        return func(*args, **kwargs)
    return decorated

class eventBase(Dataset):
    def __init__(self, 
        root:str, 
        fname:list, 
        download: bool,
        
        loader:Callable,
        subSet:str="Train", 
        extention:str=".npy",
        transform=None,
        target_transform=None,
        ) -> None:
        """
        wrapper download, and basic function like read datasets, getitem, len.
        @root: root path of the datasets(str).
        @fname: the download, extract, save folder name.
        @loader: loader data for format of extention.
        @download: whether to download datasets.
        @extention: the saved file format.
        @subSet
        @transform: transform x
        @target_transform: transform y
        @multi_process: multiprogress and Breakpoint continuingly(not implement yet).
        you need to implement downloadResource, restructureAndConvert.
        this will help you download from net and 
        restructure the datasets into specific format(extention), and the structured as follow:
        directory/
            ├── class_x
            │   ├── xxx.ext
            │   ├── xxy.ext
            │   └── ...
            │       └── xxz.ext
            └── class_y
                ├── 123.ext
                ├── nsdf3.ext
                └── ...
                └── asd932_.ext
        the basic format of sample as (t, x, y, p), label as (i).

        if you want other, or save other format in disk, please implement restructureAndConvert first,
        and reload function: make_dataset, you can save datasets in one file, or other sample format.
        if there are subsequent processing of sample data, like EST, FrameCount, etc., 
        you can reload function: __getitem__.
        ori_format is (t, x, y, p), label: (i).
        the dataflow is: download -> restructureAndConvert(ori_format) -> make_datasets -> __getitem__ -> out
        '''
    """

        self.root = root
        assert len(fname) == 2 or len(fname) == 3, f"there are {len(fname)} folder names. the download and save name must be different, and no extra."
        assert subSet in ["Train", "Test", "Validation"], 'the subSet must in ["Train", "Test", "Validation"]'
        
        self.fdownload = os.path.join(self.root, fname[0])
        self.fextract = os.path.join(self.root, fname[-2])
        self.fsave = os.path.join(self.root, fname[-1])
        self.extention = extention
        
        reload = False
        if download:
            reload = self.download(self.fdownload, self.fextract)
        if reload or not os.path.exists(self.fsave):
            self.restructureAndConvert(self.fextract, self.fsave, self.extention)
        
        self.samples = self.make_dataset(self.fsave,subSet,extensions=self.extention)
        self.loader = loader
        self.transform = transform
        self.target_transform = target_transform
        if len(self.samples) == 0:
            raise (RuntimeError("Found 0 files in subfolders of: " + os.path.join(self.fsave, subSet) + "\n"
                                "Supported extensions are: " + self.extention))

    
    @abstractmethod 
    def downloadResource(self)->list:
        """
        @return: a list of download resorce, which will be called in download function.
        or raise error when you don't know how to implement it and set download False.
        @rtype:[(str,str,str)]: [(filename, url, md5)]
        """
        raise NotImplementedError

    @abstractmethod 
    def restructureAndConvert(self, fextracted:str, fsave:str, ext:str)->None:
        """
        read from the original directory structure in source(path extracted), 
        and split into train/validate/test set(if not),
        and convert to other format and save.

        restructure the datasets into specific format(extention), and the structured as follow:
        fsave/
            ├── Train
            |   ├── class_x
            |   │   ├── xxx.ext
            |   │   ├── xxy.ext
            |   │   ├── ...
            |   │   └── xxz.ext
            |   └── class_y
            |       ├── 123.ext
            |       ├── nsdf3.ext
            |       ├── ...
            |       └── asd932_.ext
            ├── Test
            |   ├── ...
        the basic format of sample as (t, x, y, p), label as (i).
        make sure the fsave will be structured as required. and save single event file into extention file.
        read as list [t, x, y, p]
        """
        raise NotImplementedError

      
    def make_dataset(
        self,
        fsave: str,
        subSet: str,
        class_to_idx: Optional[Dict[str, int]] = None,
        extensions: Optional[Tuple[str, ...]] = None,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ) -> List[Tuple[str, int]]:
        """Generates a list of samples of a form (path_to_sample, class).

        This can be overridden to e.g. read files from a compressed zip file instead of from the disk.
        OR when you save all the datasets in one file, thus you can read one-time.

        Args:
            directory (str): root dataset directory, corresponding to ``self.root``.
            class_to_idx (Dict[str, int]): Dictionary mapping class name to class index.
            extensions (optional): A list of allowed extensions.
                Either extensions or is_valid_file should be passed. Defaults to None.
            is_valid_file (optional): A function that takes path of a file
                and checks if the file is a valid file
                (used to check of corrupt files) both extensions and
                is_valid_file should not be passed. Defaults to None.

        Raises:
            ValueError: In case ``extensions`` and ``is_valid_file`` are None or both are not None.
            FileNotFoundError: In case no valid file was found for any class.

        Returns:
            List[Tuple[str, int]]: samples of a form (path_to_sample, class)

        Note: The class_to_idx parameter is here optional and will use the logic of the ``find_classes`` function
        by default.
        """
        directory = os.path.expanduser(os.path.join(fsave, subSet))

        if class_to_idx is None:
            class_to_idx = self.find_classes_FromFolder(directory)

        if extensions is None == is_valid_file is None:
            raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")

        if extensions is not None:
            def is_valid_file(x: str) -> bool:
                return x.lower().endswith(extensions)

        instances = []
        available_classes = set()
        for target_class, class_index in class_to_idx.items():
            target_dir = os.path.join(directory, target_class)
            if not os.path.isdir(target_dir):
                continue
            for root, _, fnames in os.walk(target_dir, followlinks=True):
                for fname in fnames:
                    if is_valid_file(fname):
                        instances.append((os.path.join(root, fname), class_index))

                        if target_class not in available_classes:
                            available_classes.add(target_class)

        empty_classes = set(class_to_idx.keys()) - available_classes
        if empty_classes:
            msg = f"Found no valid file for the classes {', '.join(sorted(empty_classes))}. "
            if extensions is not None:
                msg += f"Supported extensions are: {', '.join(extensions)}"
            raise FileNotFoundError(msg)

        return instances

    def find_classes_FromFolder(self, directory: str, filter_dir:Optional[Callable]=lambda x: True) -> Dict[str, int]:
        """
        Find the class folders in a dataset structured as follows::

            directory/
            ├── class_x
            │   ├── xxx.ext
            │   ├── xxy.ext
            │   └── ...
            │       └── xxz.ext
            └── class_y
                ├── 123.ext
                ├── nsdf3.ext
                └── ...
                └── asd932_.ext

        This method can be overridden to adapt to a different dataset directory structure.
        When you need a subset or exclude some files, filter_dir may help you.

        Args:
            directory(str): Path to find. Root directory path.
            filter_dir: func(dir:os.DirEntry)->bool  a function to filter the directories of classes. return True if need

        Raises:
            FileNotFoundError: If ``dir`` has no class folders.

        Returns:
            Dict class_name : idx
        
        """
        classes = [entry.name for entry in os.scandir(directory) if entry.is_dir() and filter_dir(entry)]
        if not classes:
            raise FileNotFoundError(f"Couldn't find any class folder in {directory}.")

        class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}

        return class_to_idx

    def __getitem__(self, index):
        """
        Args:
            index (int): Index

        Returns:
            tuple: (sample, target) where target is class_index of the target class.
        """
        path, target = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return sample, target

    def __len__(self):
        return len(self.samples)

    def _download(self, root:str, fextracted:str, filename:str, url:str, md5:str)->None:
        """
        A wrapper of download_and_extract_archive
        for errors:
        it will print message if files can not be extracted, which is ValueError
        otherwise raise RuntimeError.
        """
        try:
            print(f'Download [{filename}] from [{url}] to [{root}]')
            download_and_extract_archive(url, root, fextracted, filename=filename, md5=md5)
        except ValueError as ve:
            print(ve)
        except BaseException:
            raise RuntimeError('download error, please download manually!')

    def download(self, fdownload:str, fextracted:str = None) -> bool:
        """
        If the dataset doesn't exist, download it, otherwise check for corrections.
        @return: return True if download files from internet, otherwise False.
        """
        reload = False
        if not os.path.exists(fdownload):
            os.makedirs(fdownload)
        for filename, url, md5 in self.downloadResource():
            filepath = os.path.join(fdownload, filename)
            print(f"check file:{filepath}")
            if not check_integrity(filepath, md5):
                reload = True
                if os.path.exists(filepath):
                    print(f'The file:[{filepath}] is incorrect.')
                    os.remove(filepath)
                else:
                    print(f'The file:[{filepath}] does not exist.')
                print(f'Start to download and extract:[{filepath}].')
                self._download(fdownload, fextracted, filename, url, md5)
            if not os.path.exists(os.path.join(fextracted, os.path.splitext(filename)[0])):
                print(f'Start to extract:{filepath} to {fextracted}.')
                extract_archive(filepath, fextracted)
        return reload
    
    def __repr__(self) -> str:
        head = "Dataset " + self.__class__.__name__
        body = ["Number of datapoints: {}".format(self.__len__())]
        if self.root is not None:
            body.append("Root location: {}".format(self.root))
        body += self.extra_repr().splitlines()
        lines = [head] + [" " * 4 + line for line in body]
        return '\n'.join(lines)

    def extra_repr(self) -> str:
        return ""


class cifar10_dvsBase(eventBase):
    """
    basic cifar10_dvs, will download, restructure, and read sample, label as (t, x, y, p), i.
    since cifar10_dvs is not splited into subset, split parameter will transform the ratio.

    we implement download and restructureAndConvert. so all the data get the same directory structure and format.
    for more process, reload make_datasets, __getitem__

    @root: root path of the datasets(str).
    @folders_names: the download, extract, save folder name.
    @download: whether to download datasets.

    @split: The ratio of datasets
    @subSet
    @extention: the saved file format.
    @transform: transform x
    @target_transform: transform y

    
    """
    def __init__(self,
        root:str, 
        folders_names:list=["oriDownload", "extract", "convert"],
        download: bool=False,
        split: list=[9, 1],
        subSet: str = "Train",
        extention:str=".npy",
        transform=None, 
        target_transform=None
        # saver:Callable=np.save
        ) -> None:

        self.height =128
        self.width = 128
        self.split = split
        def myloader(path)->ev.event:
            return ev.load(path, extention)
        def mysaver(path, data:ev.event):
            ev.save(path, data, extention)
        self.saver = mysaver
        super().__init__(root, folders_names, download, myloader, subSet, extention, transform, target_transform)
    
    
    def loadaerdat(self, filename)->ev.event:
        '''
        read the events
        '''
        return ev.loadaerdat(filename, camera='DVS128')

    def splitSet(self, all_img, split, fsave, classname):
        """
        split specific class's imgs.
        """
        assert len(split) > 1, f"split error:{split}, we need at least 2 number"
        
        t = float(sum(split))
        split = [math.ceil(len(all_img)/t*i) for i in split]
        split[1] += split[0]
        train_path = os.path.join(fsave, "Train", classname)
        test_path = os.path.join(fsave, "Test", classname)
        if len(split) == 2:
            print(f"split {classname} into Train and Test set...")
            train_img, test_img = np.split(np.random.permutation(all_img), 
                    [split[0]])
            return {train_path: train_img, test_path: test_img}
        elif len(split) == 3:
            print("split data into Train, Validation and Test set...")
            train_img, validation_img, test_img = np.split(np.random.permutation(all_img), 
                    [split[0], split[1]])
            validation_path = os.path.join(fsave, "Validation", classname)
            return {train_path: train_img, validation_path:validation_img, test_path: test_img}

    def restructureAndConvert(self, fextracted:str, fsave:str, ext:str)->None:
        """
        restructure the directory of datasets, since cifar10_dvs is not splited, split by yourself when restructure.
        and convert into format of ext.
        fills will convert from fextarcted to fsave.
        """
        class_to_idx = self.find_classes_FromFolder(fextracted)

        for name in class_to_idx.keys():
            ex_class_path = os.path.join(fextracted, name)
            if os.path.isdir(ex_class_path):
                all_img = os.listdir(ex_class_path)
                path_to_imgs = self.splitSet(all_img, self.split, fsave, name)
            print("start to convert ", name, "...\n")
            for targetClassPath, setImgs in path_to_imgs.items():
                for idx, img in enumerate(setImgs):
                    print(f"\033[1AImage's idx:{idx}\033[K")
                    if not os.path.exists(targetClassPath):
                        os.makedirs(targetClassPath)
                    targetImgPath = os.path.join(targetClassPath, os.path.splitext(img)[0]+ext)
                    if not os.path.exists(targetImgPath):
                        evdata = self.loadaerdat(os.path.join(ex_class_path, img))
                        self.saver(targetImgPath, evdata.removeInvalid(dim_x=self.width, dim_y=self.height))
    
    def downloadResource(self) -> list:
        '''
        :return: A list ``url`` that ``url[i]`` is a tuple, which contains the i-th file's name, download link, and MD5
        :rtype: list
        if you can not found the resource. just raise error.
        '''
        return [
            ('airplane.zip', 'https://ndownloader.figshare.com/files/7712788', '0afd5c4bf9ae06af762a77b180354fdd'),
            ('automobile.zip', 'https://ndownloader.figshare.com/files/7712791', '8438dfeba3bc970c94962d995b1b9bdd'),
            ('bird.zip', 'https://ndownloader.figshare.com/files/7712794', 'a9c207c91c55b9dc2002dc21c684d785'),
            ('cat.zip', 'https://ndownloader.figshare.com/files/7712812', '52c63c677c2b15fa5146a8daf4d56687'),
            ('deer.zip', 'https://ndownloader.figshare.com/files/7712815', 'b6bf21f6c04d21ba4e23fc3e36c8a4a3'),
            ('dog.zip', 'https://ndownloader.figshare.com/files/7712818', 'f379ebdf6703d16e0a690782e62639c3'),
            ('frog.zip', 'https://ndownloader.figshare.com/files/7712842', 'cad6ed91214b1c7388a5f6ee56d08803'),
            ('horse.zip', 'https://ndownloader.figshare.com/files/7712851', 'e7cbbf77bec584ffbf913f00e682782a'),
            ('ship.zip', 'https://ndownloader.figshare.com/files/7712836', '41c7bd7d6b251be82557c6cce9a7d5c9'),
            ('truck.zip', 'https://ndownloader.figshare.com/files/7712839', '89f3922fd147d9aeff89e76a2b0b70a7')
        ]
 

class mnist_dvsBase(eventBase):
    """
    basic cifar10_dvs, will download, restructure, and read sample, label as (t, x, y, p), i.
    since cifar10_dvs is not splited into subset, split parameter will transform the ratio.

    we implement download and restructureAndConvert. so all the data get the same directory structure and format.
    for more process, reload make_datasets, __getitem__

    @root: root path of the datasets(str).
    @folders_names: the download, extract, save folder name.
    @download: whether to download datasets.

    @split: The ratio of datasets
    @subSet
    @extention: the saved file format.
    @transform: transform x
    @target_transform: transform y

    
    """
    def __init__(self,
        root:str, 
        scale: str = "scale16",
        folders_names:list=["oriDownload", "extract", "convert"],
        download: bool=False,
        split: list=[9, 1],
        subSet: str = "Train",
        extention:str=".npy",
        transform=None, 
        target_transform=None
        # saver:Callable=np.save
        ) -> None:

        self.height =128
        self.width = 128
        self.split = split
        myloader = lambda x: ev.load(x, extention)
        def mysaver(path, data:ev.event):
            ev.save(path, data, extention)
        self.saver = mysaver
        self.scale = scale
        super().__init__(root, folders_names, download, myloader, subSet, extention, transform, target_transform)
    
    def make_dataset(self, fsave: str, subSet: str, class_to_idx: Optional[Dict[str, int]] = None, extensions: Optional[Tuple[str, ...]] = None, is_valid_file: Optional[Callable[[str], bool]] = None) -> List[Tuple[str, int]]:
        fsave = os.path.join(fsave, self.scale)
        return super().make_dataset(fsave, subSet, class_to_idx, extensions, is_valid_file)
   
    def loadaerdat(self, filename)->ev.event:
        '''
        read the event data and return (ts, xpos, ypos, pol) 4-tuple of lists containing data of all events
        '''
        return ev.loadaerdat(filename, camera = 'DVS128')

    def splitSet(self, all_img, split, fsave, classname):
        """
        split specific class's imgs.
        """
        assert len(split) > 1, f"split error:{split}, we need at least 2 number"
        
        t = float(sum(split))
        split = [math.ceil(len(all_img)/t*i) for i in split]
        split[1] += split[0]
        train_path = os.path.join(fsave, "Train", classname)
        test_path = os.path.join(fsave, "Test", classname)
        if len(split) == 2:
            print(f"split {classname} into Train and Test set...")
            train_img, test_img = np.split(np.random.permutation(all_img), 
                    [split[0]])
            return {train_path: train_img, test_path: test_img}
        elif len(split) == 3:
            print("split data into Train, Validation and Test set...")
            train_img, validation_img, test_img = np.split(np.random.permutation(all_img), 
                    [split[0], split[1]])
            validation_path = os.path.join(fsave, "Validation", classname)
            return {train_path: train_img, validation_path:validation_img, test_path: test_img}

    def restructureAndConvert(self, fextracted:str, fsave:str, ext:str)->None:
        """
        restructure the directory of datasets, since cifar10_dvs is not splited, split by yourself when restructure.
        and convert into format of ext.
        fills will convert from fextarcted to fsave.
        """
        # scale = self.scale
        # scales = self.find_classes_FromFolder(fextracted)
        scales = ['scale4','scale8','scale16']
        for scale in scales:
            fsave_scale = os.path.join(fsave, scale)
            class_to_idx = self.find_classes_FromFolder(fextracted)

            for name in class_to_idx.keys():
                ex_class_path = os.path.join(fextracted, name)
                ex_class_scale_path = os.path.join(ex_class_path, scale)
                if os.path.isdir(ex_class_scale_path):
                    all_img = os.listdir(ex_class_scale_path)
                    path_to_imgs = self.splitSet(all_img, self.split, fsave_scale, name)
                print("start to convert ", name, "...\n")
                for targetClassPath, setImgs in path_to_imgs.items():
                    for idx, img in enumerate(setImgs):
                        print(f"\033[1AImage's idx:{idx}\033[K")
                        if not os.path.exists(targetClassPath):
                            os.makedirs(targetClassPath)
                        targetImgPath = os.path.join(targetClassPath, os.path.splitext(img)[0]+ext)
                        if not os.path.exists(targetImgPath):
                            evdata = self.loadaerdat(os.path.join(ex_class_scale_path, img))
                            self.saver(targetImgPath, evdata.removeInvalid(dim_x=self.width, dim_y=self.height))


    def downloadResource(self) -> list:
        '''
        :return: A list ``url`` that ``url[i]`` is a tuple, which contains the i-th file's name, download link, and MD5
        :rtype: list
        if you can not found the resource. just raise error.
        '''
        return [
            ('grabbed_data0.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data0.zip', '451C1D2F6B195CE96D1DD478EE8AAF6E'.lower()),
            ('grabbed_data1.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data1.zip', '22259AE8871800B5EECF9E0EB8146EA5'.lower()),
            ('grabbed_data2.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data2.zip', '6D3AD96939B702F095D027CE89F1297C'.lower()),
            ('grabbed_data3.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data3.zip', 'E4833915A1514598C4AE046E5E09CF93'.lower()),
            ('grabbed_data4.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data4.zip', 'E183266A0B670B5336BA4D39D4BD30AE'.lower()),
            ('grabbed_data5.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data5.zip', 'ABF900993C268A763639C3074B6F92BB'.lower()),
            ('grabbed_data6.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data6.zip', '971DB9A99E95B408837518C7727A223C'.lower()),
            ('grabbed_data7.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data7.zip', 'A3CB115F19A736F4FC3590F36012BC87'.lower()),
            ('grabbed_data8.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data8.zip', 'FE5DA6C6F559ABF397A32439A6C1CD3F'.lower()),
            ('grabbed_data9.zip', 'http://www2.imse-cnm.csic.es/caviar/MNIST_DVS/grabbed_data9.zip', '73A3007C1C1AB4BFAD9CE07EA76D540D'.lower())
        ]


class n_carsBase(eventBase):
    """
    basic cifar10_dvs, will download, restructure, and read sample, label as (t, x, y, p), i.
    since cifar10_dvs is not splited into subset, split parameter will transform the ratio.

    we implement download and restructureAndConvert. so all the data get the same directory structure and format.
    for more process, reload make_datasets, __getitem__

    @root: root path of the datasets(str).
    @folders_names: the download, extract, save folder name.
    @download: whether to download datasets.

    @split: The ratio of datasets
    @subSet
    @extention: the saved file format.
    @transform: transform x
    @target_transform: transform y

    
    """
    def __init__(self,
        root:str, 
        folders_names:list=["oriDownload", "extract", "convert"],
        download: bool=False,
        subSet: str = "Train",
        extention:str=".npy",
        transform=None, 
        target_transform=None
        # saver:Callable=np.save
        ) -> None:

        self.height = 100
        self.width = 120
        myloader = lambda x: ev.load(x, extention)
        def mysaver(path, data:ev.event):
            ev.save(path, data, extention)
        self.saver = mysaver
        super().__init__(root, folders_names, download, myloader, subSet, extention, transform, target_transform)

    def loadaerdat(self, filename)->ev.event:
        '''
        read the event data and return (ts, xpos, ypos, pol) 4-tuple of lists containing data of all events
        '''
        return ev.loadaerdat(filename, camera='N')

    def restructureAndConvert(self, fextracted:str, fsave:str, ext:str)->None:
        """
        restructure the directory of datasets, since cifar10_dvs is not splited, split by yourself when restructure.
        and convert into format of ext.
        fills will convert from fextarcted to fsave.
        """
        fextracted = os.path.join(fextracted, "n-cars")
        test_or_train = ['Test','Train']
        for tt in test_or_train:
            fsave_tt = os.path.join(fsave, tt)
            # fextracted_class = os.path.join(fextracted, tt)
            class_to_idx = self.find_classes_FromFolder(fextracted)

            for name in class_to_idx.keys():
                ex_class_path = os.path.join(fextracted, name)
                fsave_tt_class = os.path.join(fsave_tt, name)
                if os.path.isdir(ex_class_path):
                    all_img = os.listdir(ex_class_path)
                for idx, img in enumerate(all_img):
                    print(f"\033[1AImage's idx:{idx}\033[K")
                    if not os.path.exists(fsave_tt_class):
                        os.makedirs(fsave_tt_class)
                    targetImgPath = os.path.join(fsave_tt_class, os.path.splitext(img)[0]+ext)
                    if not os.path.exists(targetImgPath):
                        evdata = self.loadaerdat(os.path.join(ex_class_path, img))
                        self.saver(targetImgPath, evdata.removeInvalid(dim_x=self.width, dim_y=self.height))
   
    def downloadResource(self) -> list:
        '''
        :return: A list ``url`` that ``url[i]`` is a tuple, which contains the i-th file's name, download link, and MD5
        :rtype: list
        if you can not found the resource. just raise error.
        '''
        return [
            # ('n-cars.zip', 'https://figshare.com/ndownloader/files/34573463', 'D1DEF7EEA897126269D19C570DF3550A'.lower())
            ('n-cars.zip', 'https://figshare.com/ndownloader/files/34573463', '914819b99dc7a6564326ba627919bf2e')
        ]



class n_caltech101Base(eventBase):
    """
    basic cifar10_dvs, will download, restructure, and read sample, label as (t, x, y, p), i.
    since cifar10_dvs is not splited into subset, split parameter will transform the ratio.

    we implement download and restructureAndConvert. so all the data get the same directory structure and format.
    for more process, reload make_datasets, __getitem__

    @root: root path of the datasets(str).
    @folders_names: the download, extract, save folder name.
    @download: whether to download datasets.

    @split: The ratio of datasets
    @subSet
    @extention: the saved file format.
    @transform: transform x
    @target_transform: transform y

    
    """
    def __init__(self,
        root:str, 
        folders_names:list=["oriDownload", "extract", "convert"],
        download: bool=False,
        split: list=[9, 1],
        subSet: str = "Train",
        extention:str=".npy",
        transform=None, 
        target_transform=None
        # saver:Callable=np.save
        ) -> None:

        self.height = 180
        self.width = 240
        self.split = split
        myloader = lambda x: ev.load(x, extention)
        def mysaver(path, data:ev.event):
            ev.save(path, data, extention)
        self.saver = mysaver
        super().__init__(root, folders_names, download, myloader, subSet, extention, transform, target_transform)

    def loadaerdat(self, filename)->ev.event:
        '''
        read the event data and return (ts, xpos, ypos, pol) 4-tuple of lists containing data of all events
        '''
        return ev.loadaerdat(filename, camera='N101')

    def splitSet(self, all_img, split, fsave, classname):
        """
        split specific class's imgs.
        """
        assert len(split) > 1, f"split error:{split}, we need at least 2 number"
        
        t = float(sum(split))
        split = [math.ceil(len(all_img)/t*i) for i in split]
        split[1] += split[0]
        train_path = os.path.join(fsave, "Train", classname)
        test_path = os.path.join(fsave, "Test", classname)
        if len(split) == 2:
            print(f"split {classname} into Train and Test set...")
            train_img, test_img = np.split(np.random.permutation(all_img), 
                    [split[0]])
            return {train_path: train_img, test_path: test_img}
        elif len(split) == 3:
            print("split data into Train, Validation and Test set...")
            train_img, validation_img, test_img = np.split(np.random.permutation(all_img), 
                    [split[0], split[1]])
            validation_path = os.path.join(fsave, "Validation", classname)
            return {train_path: train_img, validation_path:validation_img, test_path: test_img}

    def restructureAndConvert(self, fextracted:str, fsave:str, ext:str)->None:
        """
        restructure the directory of datasets, since cifar10_dvs is not splited, split by yourself when restructure.
        and convert into format of ext.
        fills will convert from fextarcted to fsave.
        """
        fextracted = os.path.join(fextracted, 'Caltech101')
        class_to_idx = self.find_classes_FromFolder(fextracted)

        for name in class_to_idx.keys():
            ex_class_path = os.path.join(fextracted, name)
            if os.path.isdir(ex_class_path):
                all_img = os.listdir(ex_class_path)
                path_to_imgs = self.splitSet(all_img, self.split, fsave, name)
            print("start to convert ", name, "...\n")
            for targetClassPath, setImgs in path_to_imgs.items():
                for idx, img in enumerate(setImgs):
                    print(f"\033[1AImage's idx:{idx}\033[K")
                    if not os.path.exists(targetClassPath):
                        os.makedirs(targetClassPath)
                    targetImgPath = os.path.join(targetClassPath, os.path.splitext(img)[0]+ext)
                    if not os.path.exists(targetImgPath):
                        evdata = self.loadaerdat(os.path.join(ex_class_path, img))
                        self.saver(targetImgPath, evdata.removeInvalid(dim_x=self.width, dim_y=self.height))
    

    def downloadResource(self) -> list:
        '''
        :return: A list ``url`` that ``url[i]`` is a tuple, which contains the i-th file's name, download link, and MD5
        :rtype: list
        if you can not found the resource. just raise error.
        '''
        return [
            ('Caltech101.zip', 'https://md-datasets-public-files-prod.s3.eu-west-1.amazonaws.com/c708f296-c555-4509-8ca9-3ff2504679ee', '66201824eabb0239c7ab992480b50ba3')
        ]
 


import torch
import numpy as np

from torch.utils.data.dataloader import default_collate


class Loader:
    def __init__(self, dataset, device, batch_size=1,num_workers=0,pin_memory=True):
        self.device = device
        split_indices = list(range(len(dataset)))
        sampler = torch.utils.data.sampler.SubsetRandomSampler(split_indices)
        self.loader = torch.utils.data.DataLoader(dataset=dataset, batch_size=batch_size, sampler=sampler,num_workers=num_workers, pin_memory=pin_memory,collate_fn=collate_events)

    def __iter__(self):
        for data in self.loader:
            data = [d.to(self.device) for d in data]
            yield data

    def __len__(self):
        return len(self.loader)


def collate_events(data):
    labels = []
    events = []
    for i, d in enumerate(data):
        labels.append(d[1])
        ev = np.concatenate([d[0], i*np.ones((len(d[0]),1), dtype=np.float32)],1)
        events.append(ev)
    events = torch.from_numpy(np.concatenate(events,0))
    labels = default_collate(labels)
    return events, labels


from collections import OrderedDict
from ctypes import Union, cast
import math
from turtle import forward
from typing import Any, List
from torchvision.models.vgg import cfgs, VGG, make_layers, vgg11, vgg13, vgg16, vgg19
from .node import Neuron
import torch.nn as nn
import torch

__all__ = ['vgg7', 'vgg11', 'vgg13', 'vgg16', 'vgg19']

cfgs['CVGG7'] = [64, 64, 'M', 128, 128, 256, 'M']
cfgs['SVGG7'] = [64, 64, 'A', 128, 128, 256]
cfgs['SVGG19'] = [64, 64, 64, 64, 64, 64, 64, 64, 'A', 128, 128, 128, 128, 128, 128, 128, 128, 256]

class cnnSmallVGG(VGG):
    def __init__(
        self,
        features: nn.Module,
        num_classes: int = 10,
    ) -> None:
        super(cnnSmallVGG, self).__init__(features, num_classes)
        self.classifier = nn.Sequential(
          nn.Linear(256 * 7 * 7, 1024),
          nn.ReLU(True),
          nn.Linear(1024, num_classes)
        )

def vgg7(pretrained: bool=False, progress: bool=True, **kwargs: Any) -> cnnSmallVGG:
    return cnnSmallVGG(make_layers(cfgs['CVGG7'], batch_norm=False), **kwargs)


def make_neurons(cfg: List, in_channels, vth:int=1, leak:int=1) -> nn.Sequential:
    layers: List[nn.Module] = []
    down_num = 0
    for v in cfg:
        if v == 'M':
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            down_num += 1
        elif v == 'A':
            layers += [Neuron(nn.AvgPool2d(kernel_size=2, stride=2), 0.75, leak, pool=True)]
            down_num += 1
        else:
            conv2d = nn.Conv2d(in_channels, v, kernel_size=3, stride=1, padding=1, bias=False)
            layers += [Neuron(conv2d, vth, leak)]
            in_channels = v
    return nn.Sequential(*layers), down_num

class snnSmallVGG(nn.Module):
    def __init__(self,
        features:nn.Sequential,
        down_num:int,
        num_classes:int=10,
        HW:tuple=(42, 42),

        vth:int=1,
        leak:int=1,
        ) -> None:
        super(snnSmallVGG, self).__init__()
        
        self.vth = vth
        self.features = features
        self.pool = nn.MaxPool2d(kernel_size=2, stride=2)
        down_size = pow(2, 1+down_num)
        self.fc1 = Neuron(nn.Linear(256 * int(HW[0]/down_size)*int(HW[1]/down_size), 1024, bias=False), vth, leak)
        self.fc2 = nn.Linear(1024, num_classes, bias=False)


    def snn_forward(self, x):
        for m in self.features:
            x = m.snn_forward(x)
        x = self.pool(x)
        x = torch.flatten(x, 1)
        x = self.fc1.snn_forward(x)
        return x
    
    def forward(self, x):
        x = self.features(x)
        x = self.pool(x)
        x = torch.flatten(x, 1)
        x = self.fc1(x)
        x = self.fc2(x)
        return x/self.vth
    
    def updateScale(self, sign):
        for m in self.features:
            m.updateScale(sign)
        self.fc1.updateScale(sign)
    
    def reset_mem(self):
        for m in self.features:
            m.reset(0,0,0)
        self.fc1.reset(0,0,0)
    

def snnvgg7(vth, leak, num_classes, CHW)->snnSmallVGG:
    return snnSmallVGG(*make_neurons(cfgs['SVGG7'], CHW[0], vth, leak), 
                    num_classes=num_classes, HW=(CHW[1], CHW[2]), vth=vth, leak=leak)

def snnvgg19(vth, leak, num_classes, CHW)->snnSmallVGG:
    return snnSmallVGG(*make_neurons(cfgs['SVGG19'], CHW[0], vth, leak), 
                    num_classes, (CHW[1], CHW[2]), vth=vth, leak=leak)


from .. import event as ev
from .fEventBase import n_caltech101Base
import torch
import numpy as np
import time

class NCALTECH101(n_caltech101Base):
    def __init__(self, 
        root: str, 
        represent:str='timesteps',
        step:int = 100,
        folders_names: list=["oriDownload", "extract", "convert"], 
        download: bool = False, 
        split: list = [9, 1], 
        subSet: str = "Train", 
        extention: str = ".npy", 
        transform=None, 
        target_transform=None) -> None:

        self.represent = represent.lower()
        self.step = step
        super().__init__(root, folders_names, download, split, subSet, extention, transform, target_transform)
        

    def __getitem__(self, index):
        path, target = self.samples[index]
        sample:ev.event = self.loader(path)
        if self.transform:
            sample = self.transform(sample)
        if self.target_transform:
            target = self.target_transform(target)
        if self.represent == 'timesteps':
            img = sample.toTimeStep((2, 180, 240, self.step), 'count')
            img = torch.from_numpy(img)
            target = torch.tensor(target)
        elif self.represent in ['est', 'eventcount', 'eventframe', 'voxgrid']:
            img = sample.toArray('xytp')
        return img, target